/* this file has been modified from the original if.c - some if.c functions are also in pb_programming_stuff.c*/
/* T Pratum 8/14/2007, modified based on spinapi from 9/4/2007 */
/* if.c
 * Functions for interfacing with RadioProcessor board
 *
 * $Date: 2007/04/20 18:25:48 $
 *
 * To get the latest version of this code, or to contact us for support, please
 * visit http://www.spincore.com
 */

/* Copyright (c) 2005 SpinCore Technologies, Inc.
 *
 * This software is provided 'as-is', without any express or implied warranty. 
 * In no event will the authors be held liable for any damages arising from the 
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 * claim that you wrote the original software. If you use this software in a
 * product, an acknowledgment in the product documentation would be appreciated
 * but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 * misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <math.h>

 #include "spinapi.h"
 #include "if.h"
 #include "caps.h"
 #include "util.h"
 #include "fid.h"
 #include "usb.h"
 
 extern char *spinerr;
 extern char *noerr;
 
 extern BOARD_INFO board[];
 extern int cur_board;

 extern double pow232;
 
 extern double last_rounded_value;
 
 
 static int set_shape_period(double period, int addr);
 
// 419 taps
// 18 bit coefficeints
// cutoff: 1/8 of nyquist (so must use decimation factor of 8 or less after fir)
//
static int fir_default2[] = {
    8122,-846,-1110,-1484,-1865,-2141,-2211,-2002,-1494,-716,240,1244,2132,
    2744,2944,2655,1872,675,-780,-2272,-3558,-4397,-4605,-4079,-2836,-1005,
    1162,3347,5181,6330,6527,5655,3733,1007,-2184,-5364,-7985,-9571,-9744,
    -8328,-5379,-1211,3631,8433,12406,14794,15011,12734,7996,1216,-6807,-14975,
    -21993,-26532,-27397,-23697,-14979,-1330,16598,37622,60130,82242,102009,
    117625,127627,131071,127627,117625,102009,82242,60130,37622,16598,-1330,
    -14979,-23697,-27397,-26532,-21993,-14975,-6807,1216,7996,12734,15011,
    14794,12406,8433,3631,-1211,-5379,-8328,-9744,-9571,-7985,-5364,-2184,
    1007,3733,5655,6527,6330,5181,3347,1162,-1005,-2836,-4079,-4605,-4397,
    -3558,-2272,-780,675,1872,2655,2944,2744,2132,1244,240,-716,-1494,-2002,
    -2211,-2141,-1865,-1484,-1110,-846,8122};
 
 
 static double shape_list[7];
 static int shape_list_offset;
 
 
 
 /**
 * \file if.c
 * \brief RadioProcessor specific functions
 *
 * This page describes only the functions which are specific to the RadioProcessor
 * product. For a complete listing of spinapi functions, please see the
 * \link spinapi.h spinapi.h \endlink documentation.
 *
 */
 


/**
 * \internal
 * Find the most significant bit that is set in the given number.
 */
int num_bits (int num)
{
	int i;
	int test = 1;
	int last_one = 0;

	for(i=0;i<32;i++) {
		if(num & test) {
			last_one = i;
		}
		test = test << 1;		
	}	
	
	return last_one + 1;	
}


/**
 * Set the number of complex points to capture. This is typically set to
 * the size of the onboard RAM, but a smaller value can be used if all
 * points are not needed.
 *
 *\param num_points The number of complex points to capture
 *
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. 0 is returned on success.
 *
 */
int pb_set_num_points(int num_points)
{
    spinerr = noerr;

    debug("pb_set_num_points: using %d points\n", num_points);

    if (num_points > board[cur_board].num_points || num_points < 0) {
        spinerr = "Number of points out of range";
        debug("pb_set_num_points: %s\n", spinerr);
        return -1;
    }
 
    reg_write (REG_SAMPLE_NUM, num_points);

    return 0;
}

/**
 * Set the number of data "segments" to be acquired. The default value is 1,
 * meaning when data is acquired, it will be stored to the RAM starting
 * at address zero, and continue until the desired number of points are
 * acquired. Any subsequent data acquisition scans will behave in the same
 * way and thus overwrite (or average with) the previous data. If num_segments
 * is set to a value higher than 1, the given number of segments will be acquired
 * before resetting the ram address to 0. For example if num_points is set to
 * 1000, and num_segments is set to 3, the first time the acquisition is triggered
 * (using scan_trigger), data will be written to RAM locations 0-999. The
 * second time it is triggered, data will be written to locations 1000-1999
 * (instead of writing again to locations 0-999 as would be the case for
 * num_segments = 1). On the third trigger data will go to locations 2000-2999.
 * Finally a fourth trigger would again write to locations 0-999, and the cycle
 * will continue as desired.
 *
 * <p>
 * When this function is called, the internal segment counter is reset to
 * segment 0.
 * </p>
 *
 *\param num_segments Number of segments to acquire. Must be between 1 and 65535.
 *
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. 0 is returned on success.
 *
 */
int pb_set_scan_segments(int num_segments)
{
    spinerr = noerr;    
    
    debug ("pb_set_scan_segments: setting number of segments to %d\n", num_segments);
    
    if (!board[cur_board].supports_scan_segments) {
        spinerr = "Your firmware revision does not support this feature";
        debug("pb_set_scan_segments: %s\n", spinerr);
        return -1;    
    }
    
    if (num_segments < 1 || num_segments > 65535) {
        spinerr = "Number of segments out of range\n";
        debug ("pb_set_scan_segments: %s\n", spinerr);        
        return -1;        
    }
    
    // reset the internal segment counter (reset is bit 16)
    reg_write (REG_SCAN_SEGMENTS, 0x010000 | (num_segments - 1) );
    // then write number of segments without the reset
    reg_write (REG_SCAN_SEGMENTS, (num_segments - 1) );
    
    return 0;
}
/**
 * Get the current value of the scan count register, or reset the register to 0. This
 * function can be used to monitor the progress of an experiment if multiple
 * scans are being performed.
 *
 * \param reset If this parameter is set to 1, this function will reset the 
 * scan counter to 0. If reset is 0, this function will return the current value of the scan counter.
 * \return The number of scans performed since the last reset is returned when reset=0. -1 is returned on error
 */
int pb_scan_count(int reset)
{
    spinerr = noerr;
    
    if (!board[cur_board].supports_scan_count) {
        spinerr = "Your firmware revision does not support this feature";
        debug("pb_scan_count: %s\n", spinerr);
        return -1;    
    }    
    
    if(reset) {
        reg_write(REG_SCAN_COUNT, 0xFF);
        reg_write(REG_SCAN_COUNT, 0x00);
        return 0;
    }
    else {
        return reg_read(REG_SCAN_COUNT);        
    }
}

/**
 * Program the onboard filters to capture data and reduce it to a baseband
 * signal with the given spectral width.
 * This function will automatically set the filter parameters and decimation
 * factors. For greater control over the filtering process, the filters can
 * be specified manually by using the pb_setup_cic() and pb_setup_fir() functions.
 *
 *\param spectral_width Desired spectral width (in MHz) of the stored baseband data. The decimation factor used is the return
 * value of this function, so that can be checked to determine the exact
 * spectral width used. If the FIR filter is used, this value must be the ADC clock divided by a multiple of 8. The value will be rounded appropriately if this condition is not met.
 *
 *\param scan_repetitions Number of scans intended to be performed. This number
 * is used only for internal rounding purposes. The actual number of scans performed
 * is determined entirely by how many times the scan_trigger control line is enabled
 * in the pulse program. However, if more scans are performed than specified here,
 * there is a chance that the values stored in RAM will overflow.
 *
 *\param cmd This paramater provides additional options for this function. 
 * Multiple options can be sent by ORing them together. If you do not wish to invoke any of the available options, use the number zero for this field. Valid options are:<ul>
 * <li>BYPASS_FIR - Incoming data will not pass through the FIR filter. This
 * eliminates the need to decimate by a multiple of 8. This is useful to obtain
 * large spetral widths, or in circumstances where the FIR is deemed unecessary.
 * Please see the RadioProcessor manual for more information about this option.
 * <li>NARROW_BW - Configure the CIC filter so that it will have a narrower bandwidth (the CIC filter will be configured to have three stages rather than the default of one).  Please see your board's product manual for more specific information on this feature.
 * </ul>
 *
 *
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. The overall decimation factor used is returned
 *on success.
 */

int pb_setup_filters(double spectral_width, int scan_repetitions, int cmd)
{
	spinerr = noerr;
	
	debug("pb_setup_filters: spectral_width %f, repetitions %d\n", spectral_width, scan_repetitions);
	
	int *coef;

    int bypass_fir = cmd & BYPASS_FIR;
    int use_narrow_bw = cmd & NARROW_BW;

    int fir_dec_amount;
    if (bypass_fir) {
        fir_dec_amount = 1;
        pb_set_radio_control(BYPASS_FIR);
        debug ("pb_setup_filters: bypassing FIR filter\n");
    }
    else {
    	fir_dec_amount = 8;
        pb_unset_radio_control(BYPASS_FIR);
    }

    int cic_stages;   // Number of stages in cic filter
    if(use_narrow_bw)
    {
        cic_stages = 3;
        debug ("pb_setup_filters: Using narrow CIC bandwidth option\n");
    }
    else{
         cic_stages = 1;
    }
    
    // Set CIC decimation amount based on the FIR dec. amount to achieve desired SW
	int cic_dec_amount = board[cur_board].clock*1000.0/(spectral_width * (double)fir_dec_amount); // x 1000 because clock is internally stored as GHz
	debug("pb_setup_filters: cic_dec_amount %d\n", cic_dec_amount);

    // filter parameters	
    int fir_shift_amount; // Will hold amount to shift right the output of the fir filter
    int cic_shift_amount; // Will hold amount to shift right the output of the cic filter
    
    int cic_m = 1;        // M value for cic filter

	// Calculate the maximum bit growth due to averaging 
	int average_bit_growth = (int)(ceil(log((double)scan_repetitions)/log(2.0)));
    /* add 2 to avg bit growth to account for overflow due to increased amp */
    average_bit_growth+=1;
    debug("pb_setup_filters: average_bit_growth %d\n", average_bit_growth);

    // Figure out what is the worst case bit growth inside the CIC filter
    int cic_bit_growth = (int)(ceil((double)cic_stages * log((double)(cic_m * cic_dec_amount))/log(2.0)));
    debug ("pb_setup_filters: cic bit growth is %d\n", cic_bit_growth);
    //CIC has 28 bit input, 35 bit output
    cic_shift_amount = cic_bit_growth + 28 - 35;
    // If bypassing the FIR filter, need to account for bit growth of averaging here
    if(bypass_fir) {
        debug("pb_setup_filters: taking care of average bit growth after CIC filter\n");
        cic_shift_amount += average_bit_growth;
    }
    if (cic_shift_amount < 0) {
        cic_shift_amount = 0;
    }
    debug("pb_setup_filters: cic_shift_amount %d\n", cic_shift_amount);

    // use builtin fir, which has a bit growth of 21 and 131 taps
    //coef = fir_default1;
	//int fir_bit_growth = 22;
	int num_taps = 131;
    int fir_bit_growth = 21;
    coef = fir_default2;

	// FIR has 35 bit input, 32 bit output
    fir_shift_amount = 35 + fir_bit_growth - 32;
    if(!bypass_fir) {
        debug("pb_setup_filters: taking care of average bit growth after FIR filter\n");
        fir_shift_amount += average_bit_growth;
    }
    if (fir_shift_amount < 0) {
        fir_shift_amount = 0;
    }
    debug("pb_setup_filters: fir_shift_amount %d\n", fir_shift_amount);

    // The FIR filter requires num_taps/2 + 4 clock cycles to process each
    // point of data. Because of this, the CIC filter must decimate by at least
    // the much or the FIR filter will not be able to keep up.
    int min_dec_amount = num_taps/2 + 4;
    if (cic_dec_amount < min_dec_amount && !bypass_fir) {
        debug ("pb_setup_filters: limiting cic_dec_amount to a minimum of %d (was %d)\n", min_dec_amount, cic_dec_amount);
        cic_dec_amount = min_dec_amount;
    }
    // 8 is the minimum decimation no matter what
    if (cic_dec_amount < 8) {
        debug ("pb_setup_filters: limiting cic_dec_amount to 8 (was %d)\n", cic_dec_amount);
        cic_dec_amount = 8;
    }

    // Pass the actual parameters to the board	
    if(pb_setup_cic (cic_dec_amount, cic_shift_amount, cic_m, cic_stages) < 0) {
        return -1;
    }
    if(pb_setup_fir (num_taps, coef, fir_shift_amount, fir_dec_amount) < 0) {
        return -1;
    }
    
	return fir_dec_amount*cic_dec_amount;
}


/**
 * Clear the contents of the data acquisition RAM to all zeros (RadioProcessor PCI boards
 * only). The function will complete in approximately 1 ms. This will not affect the RAM which stores
 * the Pulse Program or DDS information.
 *
 * <p>
 * On PCI version of RadioProcessor boards, this should always be done before an experiment is run. This is because
 * the acquisition storage system always adds the incoming data to the data
 * currently in RAM before storing it, so the RAM must be cleared before
 * the first scan of an experiment to avoid unwanted averaging.
 * </p>
 * <p>On the USB version of RadioProcessor boards, the first scan of an experiment is NOT averaged,
 * and therefore this function is not needed. pb_zero_ram() has no effect on on RadioProcessor USB boards
 * and returns immediately with an error.
 * </p>
 *
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. 0 is returned on success.
 */
int pb_zero_ram(void)
{
    spinerr = noerr;

    if(!board[cur_board].is_radioprocessor) {
        spinerr = "This function is used with RadioProcessor boards only";
        debug("pb_zero_ram: %s\n", spinerr);
        return -1;
    }

    if(board[cur_board].is_usb) {
        spinerr = "This function does nothing on your revision of the RadioProcessor";
        debug("pb_zero_ram: %s\n", spinerr);
        return -1;

// If you did actually want to zero the ram though, this is what you would do:
//        int size = 256*1024*8;
//        int i;
//        char *write_buf;
//            
//        write_buf = calloc(size);
//        if(!write_buf) {
//            debug("pb_zero_ram: write buf allocation failed\n");
//            return -1;
//        }
//
//        pb_set_radio_control(0x02);
//        usb_write_ram(BANK_DATARAM, 0, size, write_buf);
//        pb_unset_radio_control(0x02);
//
//        free(write_buf);
    }

    int old_control = reg_read(REG_CONTROL);
    
    reg_write (REG_CONTROL, old_control | DO_ZERO);
    
    // do zero must be on long enough to cycle through all memory addresses
    // For 16k data points, at 80MHz clock, this takes approx. 0.2 ms
    pb_sleep_ms(1);
        
    reg_write (REG_CONTROL, old_control);

    return 0;
}

/**
 * Retrieve the captured data from the board's memory. Data is returned as a
 * signed 32 bit integer. Data can be accessed at any time, even while the data
 * from a scan is being captured. However, this is not recommened since there
 * is no way to tell what data is part of the current scan and what is part
 * of the previous scan.<br>
 * pb_read_status() can be used to determine whether or not a scan is currently
 * in progress.<br>
 * It takes approximately 160ms to transfer all 16k complex points.
 *
 *\param num_points Number of complex points to read from RAM
 *\param real_data Real data from RAM is stored into this array
 *\param imag_data Imag data from RAM is stored into this array
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_get_data(int num_points, int *real_data, int *imag_data)
{
	int i, ret;
    int control;
    int tmp[2*16*1024];
    int board_points;
    int pos;

    spinerr = noerr;

    // This holds the number of points that can be stored on the board
    board_points = board[cur_board].num_points;

	if (num_points > board_points) {
		spinerr = "Too many points";
		debug ("pb_get_data: %s (%d > %d)\n", spinerr, num_points, board[cur_board].num_points);
		return -1;
	}
    if(num_points < 1) {
        spinerr = "num_points must be > 0";
		debug ("pb_get_data: %s\n", spinerr);
        return -1;
    }

    if(board[cur_board].is_usb) {
        debug("pb_get_data: getting data from board over USB protocol\n");
        // FIXME: we could get around this limitation by fixing the code in usb_read_ram()
        if (num_points % (8*1024) != 0) {
            spinerr = "Data transfers must be a multiple of 8k points";
            debug("pb_get_data %s\n", spinerr);
            return -1;
        }

        pb_set_radio_control(0x02); // turn on the PCI_READ bit

        char *read_buf;
        read_buf = malloc(num_points*8);
        if(!read_buf) {
            spinerr = "Internal error: can't allocate read buffer";
            debug("pb_get_data: %s\n", spinerr);
            return -1;
        }

        /* note that read_ram returns 0 on success */
        ret = usb_read_ram(BANK_DATARAM, 0, num_points*8, read_buf);
        if(ret >= 0)
        {
        
        	for(i = 0; i < num_points*8; i+= 8) {
            	real_data[i/8] = (0x0FF & read_buf[i]);
            	real_data[i/8] |= (0x0FF & read_buf[i+1])<<8;
            	real_data[i/8] |= (0x0FF & read_buf[i+2])<<16;
            	real_data[i/8] |= (0x0FF & read_buf[i+3])<<24;

            	imag_data[i/8] = (0x0FF & read_buf[i+4]);
            	imag_data[i/8] |= (0x0FF & read_buf[i+5])<<8;
            	imag_data[i/8] |= (0x0FF & read_buf[i+6])<<16;
            	imag_data[i/8] |= (0x0FF & read_buf[i+7])<<24;                        
       		 }

       	}
       	else
        	{
            spinerr = "USB ram reading failed";
            /*printf("pb_get_data: %s, %d\n", spinerr,ret);*/
            debug("pb_get_data: %s\n", spinerr);
			}        

        free(read_buf);
 
        pb_unset_radio_control(0x02); // turn off the PCI_READ bit

        return ret;
    }


    //debug("pb_get_data: mem address (start): %d\n", pb_inw(MEM_ADDRESS));

    control = reg_read(REG_CONTROL);
	// The PCI_READ control bit must be set to be able to read data from RAM
    reg_write (REG_CONTROL, control | PCI_READ);

    // this is a fix for the "wraparound" issue in the 10-4 firmware (and 10-5, 10-6)
    // the problem is that writing to the address register does not work (but reading does), so
    // there is no way to specify where to start reading in the RAM. so we have
    // to just read it all, and then shift appropriately based on reading the value
    // from the address register
    #define F4_RSIZE (2*16*1024) // 16k for each channel
    if(board[cur_board].has_wraparound_problem) {
        debug("pb_get_data: using wraparound fix\n");

        pos = pb_inw(MEM_ADDRESS);
        // read in all data. tmp will hold data in the correct order
    	for (i = 0; i<F4_RSIZE; i++) {
            while(pos >= F4_RSIZE) {
                pos -= F4_RSIZE;
            }
    	    tmp[pos] = pb_inw (MEM_DATA);
            pos++;
    	}
        for (i = 0; i<num_points; i++) {
            real_data[i] = tmp[2*i];
            imag_data[i] = tmp[2*i+1];
        }
    }
    // Otherwise just read ram in the normal way
    else {
    	// Reset memory address register
    	pb_outw (MEM_ADDRESS, 0);

    	//loop num_points times and read values from ram
    	for (i = 0; i<num_points;i++)
    	{
    	  real_data[i] = pb_inw (MEM_DATA);
          imag_data[i] = pb_inw (MEM_DATA);
    	}
    }



    reg_write (REG_CONTROL, control);

    //debug("pb_get_data: mem address (end): %d\n", pb_inw(MEM_ADDRESS));

	return 0;
}

/**
 * Retrieve captured data from the board's memory. Use this function instead of pb_get_data()
 * if you have used the direct capture (data points are sent from the A/D directly to RAM with
 * no modification) capabilities of the board. Direct capture points are the raw 14 bit, signed
 * data from the A/D and are returned as an array of 16 bit short integers.<br>
 * You can enable direct capture by calling the function pb_set_radio_control(RAM_DIRECT)<br>
 * In direct capture mode, the board will acquire 4x the number of points you set with pb_set_num_points()
 * 
 * \param num_points Number of direct capture points to retrieve
 * \param data Storage space for the data points. This must contain enough room
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */

int pb_get_data_direct(int num_points, short *data)
{
    spinerr = noerr;

    int i;

    if(board[cur_board].is_usb) {
        debug("pb_get_data_direct: getting data from board via USB protocol\n");

        // FIXME: we could get around this limitation by fixing the code in usb_read_ram()
        if (num_points % (32*1024) != 0) {
            spinerr = "Direct data transfers must be a multiple of 32k points";
            debug("pb_get_data %s\n", spinerr);
            return -1;
        }

        pb_set_radio_control(0x02); // turn on the PCI_READ bit

        char *read_buf = malloc(num_points*2);
        if(!read_buf) {
            spinerr = "Internal error, malloc() failed";
            debug("pb_set_debug: %s\n");
            return -1;
        }

        usb_read_ram(BANK_DATARAM, 0, num_points*2, read_buf);
        
        for(i = 0; i < num_points*2; i+= 2) {
            data[i/2] = (0x0FF & read_buf[i]);
            data[i/2] |= (0x0FF & read_buf[i+1])<<8;
        }
 
        int a, b, c, d;
        // rotate each block of 4 around by 1
        for(i=0; i < num_points; i+= 4) {
            a = data[i];
            b = data[i+1];
            c = data[i+2];
            d = data[i+3];
        
            data[i]   = b;
            data[i+1] = c;
            data[i+2] = d;
            data[i+3] = a;
        }

 
 
        free(read_buf);
 
        pb_unset_radio_control(0x02); // turn off PCI_READ bit

        // FIXME: This was present for debugging purposes. It can probably removed now
        int q = reg_read(REG_OVERFLOW_COUNT);
        //printf("capture_ram_address is 0x%x (%d)\n", a, a);

        q = reg_read(REG_OVERFLOW2_COUNT);
        //printf("segment_counter is 0x%x (%d)\n", a, a);
    }    

    return 0;    
}

/** Deprecated legacy function.  Please use pb_write_ascii_verbose instead. */
int pb_write_ascii(char * fname, int num_points, float SW, int *real_data, int *imag_data)
{
		debug ("pb_write_ascii: Spectrometer Frequency will be written to output file as zero (use pb_write_ascii_verbose to avoid this).\n");

		int i;
		i = pb_write_ascii_verbose(fname,num_points,SW, 0.0,real_data,imag_data);
		return (i);
}

/**
 * Write the data captured from RAM to an ascii file. The file format produced is:
 * The first three lines are comments containing information about the RadioProcessor and SpinAPI.
 * The fourth line contains the number of complex points, the fifth line
 * contains the spectrometer frequency (in MHz), the sixth line contains the spectral width 
 * of the data (in Hz), and the remaining lines
 * contain the complex points themselves. Real and Imaginary compoents of the
 * complex points are given on alternate lines. Thus, the real and imaginary
 * components of the first point are given on lines 7 and 8 respectively. The
 * second point is given on lines9 and 10, etc.
 *
 *\param fname Filename to write the data to
 *\param num_points Number of complex data points to write
 *\param SW Spectral width in Hz. This should be set to the spectral width of the stored baseband data.
  *\param SF Spectrometer frequency in MHz
 *\param real_data Array holding the real portion of the complex data points
 *\param imag_data Array holding the imaginary portion of the complex data points
 *\return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_write_ascii_verbose (char * fname, int num_points, float SW, float SF, int *real_data, int *imag_data)
{
	int i;

    spinerr = noerr;

	//try to open file
	FILE * fp = fopen(fname, "w");
	if (fp == NULL) {
        spinerr = "Couldnt open file";
	   debug("pb_write_ascii: %s\n", spinerr);
		return -1;
    }

	//write file tag info
	fprintf(fp, "#SpinCore Technologies, Inc.\n#SF in MHz\n#SW in Hz\n\n");
	fprintf(fp, "[HEADER]\n");
	fprintf(fp, "@CLASS=	RADIOPROCESSOR\n");
	fprintf(fp, "@FIRMWARE= %d\n",board[cur_board].firmware_id);
	fprintf(fp, "@SPINAPI= %s\n",pb_get_version());
	fprintf(fp, "@POINTS= %d\n",num_points);	
	fprintf(fp, "@SF= %f\n",SF);	
	fprintf(fp, "@SW= %f\n\n",SW);	
	fprintf(fp, "[DATA]\n");
	
	for (i=0; i < num_points; i++) {
		fprintf (fp, "%d\n", real_data[i]);
		fprintf (fp, "%d\n", imag_data[i]);		
	}

	fclose(fp);

	return 0;
}

int pb_write_datafile (char * fname, int num_points, float SW, float SF, int *real_data, int *imag_data)
{
	/* THIS WRITES A BINARY FILE use the routine above to write an ascii file - otherwise the formats are the same */
	int i,j;
	int *udata = (int *)malloc(2*num_points*sizeof(int));

    spinerr = noerr;

	//try to open file
	FILE * fp = fopen(fname, "w");
	if (fp == NULL) {
        spinerr = "Couldnt open file";
	   debug("pb_write_ascii: %s\n", spinerr);
		return -1;
    }

	//write file tag info
	fprintf(fp, "#SpinCore Technologies, Inc.\n#SF in MHz\n#SW in Hz\n\n");
	fprintf(fp, "[HEADER]\n");
	fprintf(fp, "@CLASS=	RADIOPROCESSOR\n");
	fprintf(fp, "@FIRMWARE= %d\n",board[cur_board].firmware_id);
	fprintf(fp, "@SPINAPI= %s\n",pb_get_version());
	fprintf(fp, "@POINTS= %d\n",num_points);	
	fprintf(fp, "@SF= %f\n",SF);	
	fprintf(fp, "@SW= %f\n\n",SW);	
	fprintf(fp, "[DATA]\n");
	
	j = 0;
	/*load complex data array*/
	for (i=0; i < num_points; i++) {
		udata[j++] = real_data[i];
		udata[j++] = imag_data[i];
	}
	/*write complex data array record*/
	fwrite(udata,sizeof(int),2*num_points,fp);

	fclose(fp);

	return 0;
}


int pb_append_datafile (char * fname, int num_points, float SW, float SF, int *real_data, int *imag_data)
{
	/* THIS WRITES A BINARY FILE use the routine above to write an ascii file - otherwise the formats are the same; this routine appends the data to an existing file */
	int i,j;
	int *udata = (int *)malloc(2*num_points*sizeof(int));

    spinerr = noerr;

	//try to open file
	FILE * fp = fopen(fname, "a");
	if (fp == NULL) {
        spinerr = "Couldnt open file";
	   debug("pb_write_ascii: %s\n", spinerr);
		return -1;
    }

	//write file tag info
	fprintf(fp, "#SpinCore Technologies, Inc.\n#SF in MHz\n#SW in Hz\n\n");
	fprintf(fp, "[HEADER]\n");
	fprintf(fp, "@CLASS=	RADIOPROCESSOR\n");
	fprintf(fp, "@FIRMWARE= %d\n",board[cur_board].firmware_id);
	fprintf(fp, "@SPINAPI= %s\n",pb_get_version());
	fprintf(fp, "@POINTS= %d\n",num_points);	
	fprintf(fp, "@SF= %f\n",SF);	
	fprintf(fp, "@SW= %f\n\n",SW);	
	fprintf(fp, "[DATA]\n");
	
	j = 0;
	/*load complex data array*/
	for (i=0; i < num_points; i++) {
		udata[j++] = real_data[i];
		udata[j++] = imag_data[i];
	}
	/*write complex data array record*/
	fwrite(udata,sizeof(int),2*num_points,fp);

	fclose(fp);

	return 0;
}


/**
 * Write the RAM contents to a JCAMP-DX file. 
 *\param fname The filename for the file you want to create
 *\param num_points Number of points to write to the file
 *\param SW Spectral width of the baseband data in Hz
 *\param SF Spectrometer frequency in MHz
 *\param real_data Integer array containing the real portion of the data points
 *\param imag_data Integer array containing the imaginary portion of the data points
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. 0 is returned on success.
 */
/**
 * Write the RAM contents to a JCAMP-DX file. 
 *\param fname The filename for the file you want to create
 *\param num_points Number of points to write to the file
 *\param SW Spectral width of the baseband data in Hz
 *\param SF Spectrometer frequency in MHz
 *\param real_data Integer array containing the real portion of the data points
 *\param imag_data Integer array containing the imaginary portion of the data points
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. 0 is returned on success.
 */
int pb_write_jcamp (char * fname, int num_points, float SW, float SF, int *real_data, int *imag_data)
{

    
    const float points_per_line = 6.0f;
	int i;
	int sevenPoints;
	float tmp_delay;
	float tmp_time;
	float inc;
	float time_accumulator = 0.0f;
	
	
	int min_real = real_data[0];
	int min_imag = imag_data[0];
	
	int max_real = real_data[0];
	int max_imag = imag_data[0];
	
	for(i=1; i<num_points; ++i)
	{
             if(min_real > real_data[i])
                min_real = real_data[i];
             if(max_real < real_data[i])
                max_real = real_data[i];
                
             if(min_imag > imag_data[i])
                min_imag = imag_data[i];
             if(max_imag < imag_data[i])
                max_imag = imag_data[i]; 
    }
	
    spinerr = noerr;

	//try to open file
	FILE * fp = fopen(fname, "w");
	if (fp == NULL) {
       spinerr = "Couldnt open file";
		debug("pb_write_jcamp: %s\n", spinerr);
		return -1;
    }

	tmp_delay=(float)((1/SW)*1000000);  	//Time between points in seconds
	tmp_time=(float)((1/(SW))*num_points);	// Acquisition Time in seconds
	inc = tmp_time*1000000/(float)(num_points-1);

	fprintf(fp, "##TITLE= \n");				
	fprintf(fp, "##JCAMP-DX= 5.00 $$Spinapi version %s\n", "pb_get_version()" );
	fprintf(fp, "##DATA TYPE= NMR FID\n");
	fprintf(fp, "##DATA CLASS= NTUPLES\n");
	fprintf(fp, "##ORIGIN=\n");
	fprintf(fp, "##OWNER=\n");
	fprintf(fp, "##.OBSERVE FREQUENCY=%f\n",SF); 
	fprintf(fp, "##.OBSERVE NUCLEUS=\n");
	fprintf(fp, "##.DELAY= (0,0)\n");  			//these should be the pre-acquisition delays
	fprintf(fp, "##.ACQUISITION MODE= SIMULTANEOUS\n");
	fprintf(fp, "##.DIGITISER RES=14\n");
	fprintf(fp, "##SPECTROMETER/DATA SYSTEM= SpinCore Technologies, Inc.  Radio Processor\n");
	fprintf(fp, "##NTUPLES= NMR FID\n");
	fprintf(fp, "##VAR_NAME=	TIME,		FID/REAL,	FID/IMAG,	PAGE NUMBER\n");
	fprintf(fp, "##SYMBOL=	X,		R,		I,		N\n");
	fprintf(fp, "##VAR_TYPE=	INDEPENDENT,	DEPENDENT,	DEPENDENT,	PAGE\n");
	fprintf(fp, "##VAR_FORM=	AFFN,		AFFN,		AFFN,		AFFN\n");
	fprintf(fp, "##VAR_DIM=	%u,		%u,		%u,		2\n",num_points,num_points,num_points);
	fprintf(fp, "##UNITS=	SECONDS,	ARBITRARY UNITS,ARBITRARY UNITS,\n");
	fprintf(fp, "##FIRST=	0.0,		%d,		%d,		1\n",real_data[0],imag_data[0]);
	fprintf(fp, "##LAST=	 %f,		%d,		%d,		2\n",tmp_time,real_data[num_points-1],imag_data[num_points-1]);
    fprintf(fp, "##MIN=     0.0,        %d,     %d,     1\n",min_real,min_imag);
    fprintf(fp, "##MAX=      %f,        %d,     %d,     1\n",tmp_time,max_real,max_imag);    
	fprintf(fp, "##FACTOR=	%.4E,		1,		1,		1\n",inc);

	//start real data
	fprintf(fp, "##PAGE=		N=1\n");
	fprintf(fp, "##DATA TABLE=	(X++(R..R)), XYDATA $$Real data points	\n");
	//do  points on a line until finished real_data

	for (i=0; i < num_points;) {

          fprintf (fp, "%d " ,i);
        
        for(sevenPoints = 0; sevenPoints<(int)points_per_line && i<num_points ; sevenPoints++)
		                fprintf (fp, "%d ", real_data[i++]);
        
        time_accumulator += points_per_line*inc;
		fprintf( fp, "\n");
		
	}
    time_accumulator = 0.0f;
    
	//start imag data
	fprintf(fp, "##PAGE=		N=2\n");
	fprintf(fp, "##DATA TABLE=	(X++(I..I)), XYDATA $$Imaginary data points\n");
	

     for (i=0; i < num_points;) {

        fprintf (fp, "%d " , i);
       
        for(sevenPoints = 0; sevenPoints<(int)points_per_line && i<num_points; sevenPoints++)
		                fprintf (fp, "%d ", imag_data[i++]);
        
        time_accumulator += points_per_line*inc;
		fprintf( fp, "\n");
		
	}
	
	fprintf(fp, "##END NTUPLES=	NMR FID\n");
	fprintf(fp, "##END=\n");
	
	fclose(fp);

	return 0;
}

/**
 * Write the RAM contents to a Felix file. 
 *\param fnameout The filename for the Felix file you want to create
 *\param num_points Number of points to write to the file
 *\param SW Spectral width of the baseband data in Hz
 *\param SF Spectrometer frequency in MHz
 *\param real_data Integer array containing the real portion of the data points
 *\param imag_data Integer array containing the imaginary portion of the data points
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. 0 is returned on success.
 */
int pb_write_felix(char * fnameout, int num_points, float SW, float SF, int *real_data, int *imag_data)
{
    spinerr = noerr;

	int NP = num_points;

    float mySW = SW / 1000.0;

    float spectrometer_freq = SF;

	//try to open fnameout for writing
	FILE *pFile = fopen(fnameout, "wb");
	
	//make sure fopen worked
	if (pFile == NULL) {
        spinerr = "Couldnt open output file\n";
        debug("pb_write_felix: %s\n", spinerr);
		return -1;
    }

	//the following code was mostly cut and paste from spinwrap.dll source

	long paramsize=32;
	long tmp_long;
	float tmp_fl;
	short int tmp_unit;
	bool tmp_bool;
	short int tmp_ft;
	short int tmp_dm;

	/*******************************************
	 * Marker
	 * First 32 bytes of file
	 *******************************************/
	char felixForWindows [32] = "Felix for Windows             \0";	
	int index = 0;
	unsigned int UUINT=0x01000000;

	for(index=0; index < 32-sizeof(unsigned int); index++) {
		if (felixForWindows[index] == 0x20 && index > 9)
			fputc(0x00, pFile);
		else
			fputc(felixForWindows[index], pFile);
	}
	//UINT 0x0100
	fwrite(&UUINT,sizeof(unsigned int),1,pFile);


	/*********************************************
	 * parameters: 1024 bytes - Marker (32 bytes)
	 *********************************************/
	
	//Acqusition Parameters
	tmp_ft=FID_TYPE;		// Type of FID data
	fwrite(&tmp_ft,sizeof(short),1,pFile);
	paramsize+=sizeof(short);

	tmp_long=NP;//Number of Points
	fwrite(&tmp_long,sizeof(long),1,pFile);
	paramsize+=sizeof(long);
	
	tmp_fl=mySW*(float)1000;			// Spectral Width
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);

	tmp_fl=(float)(1/(mySW*1000))*NP;	// Acquisition Time
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);

	tmp_fl=(float)spectrometer_freq;			// Spectrometer Frequency
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);

	//Processing Parameters
	tmp_long=DC;	// Drift Correction Points
	fwrite(&tmp_long, sizeof(long), 1, pFile);
	paramsize+=sizeof(long);
	tmp_long=FOURIER_NUMBER;		// Fourier Number
	fwrite(&tmp_long, sizeof(long), 1, pFile);
	paramsize+=sizeof(long);
	tmp_fl=LINE_BROAD;			// Line Broadening
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
	tmp_fl=GAUSS_BROAD;			// Gaussian Broadening
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);

	//Display Parameters
	tmp_dm=DISP_MODE;			// Display Mode
	fwrite(&tmp_dm, sizeof(short), 1, pFile);
	paramsize+=sizeof(short);
	tmp_fl=PH0;			// Zero order phase
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
	tmp_fl=PH1;			// first order phase
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
	tmp_fl=REF_LINE_FREQ;			//reference line frequency
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
    // Set the reference point for the zero frequency line. If this is
    // set to half of the number of points in the FFT, the zero frequency
    // will be in the center
	tmp_long=num_points/2;
	fwrite(&tmp_long, sizeof(long), 1, pFile);
	paramsize+=sizeof(long);

	tmp_unit=FID_UNITS;		// FID units
	fwrite(&tmp_unit,sizeof(short),1,pFile);
	paramsize+=sizeof(short);
	tmp_unit=SPEC_UNITS;		// Spectrum Units
	fwrite(&tmp_unit,sizeof(short),1,pFile);
	paramsize+=sizeof(short);
	tmp_bool=AXIS;		// Axis on/off
	fwrite(&tmp_bool,sizeof(bool),1,pFile);
	paramsize+=sizeof(bool);

	putc(0x00, pFile);				// Extra blank character ?
	paramsize+=1;

	tmp_fl=SCALE_OFFSET;			// Scale offset
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
	tmp_long=START_PLOT;		// Start of plot
	fwrite(&tmp_long, sizeof(long), 1, pFile);
	paramsize+=sizeof(long);
	tmp_long=WIDTH_PLOT;		// Width of plot
	fwrite(&tmp_long, sizeof(long), 1, pFile);
	paramsize+=sizeof(long);
	tmp_fl=VERT_OFFSET;			// Vertical offset
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
	tmp_fl=VERT_SCALE;			// vertical scale
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
	tmp_long=FID_START;		// FID start of plot
	fwrite(&tmp_long, sizeof(long), 1, pFile);
	paramsize+=sizeof(long);
	tmp_long=NP;		// FID width of plot
	fwrite(&tmp_long, sizeof(long), 1, pFile);
	paramsize+=sizeof(long);
	tmp_fl=FID_VERT_OFFSET;			// FID vertical offset
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);
	tmp_fl=FID_VERT_SCALE;			// FID vertical scale
	fwrite(&tmp_fl, sizeof(float), 1, pFile);
	paramsize+=sizeof(float);

	tmp_long=0x00000000;
	for(index=0; index<1024-paramsize; index++) {
		fwrite(&tmp_long,1,1,pFile);
	}

	/**************************
	 * Title Block: 512 bytes *
	 **************************/
	/*
	int t = 00;
	int aa = 0;
	for (aa = 0; aa < 128; aa++)
		fwrite(&t, sizeof(t), 1, pFile);
	*/
	
	char buffer[512];
	sprintf(buffer, "RadioProcessor\r\nSpinCore Technologies, Inc.\r\nwww.spincore.com\r\nNumber of Points = %d\r\nSpectral Width = %f\r\nSpectrometer Frequency = %f", num_points, SW, SF);
	fwrite(buffer, sizeof(char), 512, pFile);
	
	int k;
	long temp;
	for (k = 0; k < NP; k++)
	{	
		//get real value from ram file
		temp = real_data[k];
        //fscanf(pRam, "%ld", &temp);

		//write to felix file
		//if (k != 0)
		fwrite(&temp,sizeof(long),1,pFile);
	
		
		//get imag value from ram file
		//fscanf(pRam, "%ld", &temp);
        temp = imag_data[k];

		//write to felix file
		//if (k != 0)
		fwrite(&temp,sizeof(long),1,pFile);
	}

	fwrite(&temp,sizeof(long),1,pFile);
	fwrite(&temp,sizeof(long),1,pFile);

	//fwrite(&temp,sizeof(long),1,pFile);
	//fwrite(&temp,sizeof(long),1,pFile);


	//close the files
	if (pFile != NULL)
		fclose(pFile);
	

	return 0;
}

/**
 * Set the parameters on the onboard CIC filter. If
 * the pb_setup_filters() function is used, filter specification is done
 * automatically and this function is not necessary.
 *
 *\param dec_amount The amount of decimation the filter should perform. This can
 * be between 8 and 65535
 *\param shift_amount Amount to shift the output of the CIC filter to the right
 *\param m M parameter (differential delay) for the CIC filter. This can be 1 or 2.
 *\param stages Number of stages of the filter (1, 2, or 3)
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_setup_cic (int dec_amount, int shift_amount, int m, int stages)
{
    unsigned int word;
    unsigned int ret;

    spinerr = noerr;

    debug("pb_set_cic: dec_amount=%d shift_amount=%d m=%d stages=%d\n", dec_amount, shift_amount, m, stages);

    if (dec_amount > board[cur_board].cic_max_decim || dec_amount < 8) {
        spinerr = "dec_amount out of range";        
		debug("pb_set_cic: %s\n", spinerr);
		return -1;
    }
    
    if (shift_amount > board[cur_board].cic_max_shift || shift_amount < 0) {
        spinerr = "shift_amount out of range";        
		debug("pb_set_cic: %s\n", spinerr);
		return -1;
    }

    if (m < 1 || m > 2) {
        spinerr = "m must be 1 or 2";                
		debug("pb_set_cic: %s\n", spinerr);
		return -1;
    }
    
    if (stages < 1 || stages > board[cur_board].cic_max_stages) {
        spinerr = "stages out of range";
		debug("pb_set_cic: %s\n", spinerr);
		return -1;
    }
 
    // subtract 1 from stages and dec_amount, because hardware accepts values
    // in this way
    
    word = (stages - 1) + ((m - 1) << 12) + (shift_amount << 4) + ((dec_amount - 1) << 16);
    debug ("pb_set_cic: cic word is %x\n", word);
    
    reg_write (REG_CIC_CONTROL, word);
    
    ret = reg_read (REG_CIC_CONTROL);
    
	return 0;
    
}

/**
 * Load the coefficients for the FIR filter. This function will read floating point
 * values, one on each line, into the coef array. The coeficients will be scaled
 * appropriately to make use of the entire word coefficient word size. The coefficients
 * MUST be even symmetric.<br><br>
 * This function also calculates the worst case gain of
 * the filter. Thus the absolute largest value needed to represent the output of
 * the filter is the input word with + the worst case gain.
 * <br><br>
 * This function only fills the coef array with the coefficients given in
 * the file. To actually set these values to the board, use the pb_setup_fir()
 * function.
 * 
 *
 *\param coef Integer array that will hold the coefficients. This should have
 * enough space allocated to fit num_taps coefficients
 *\param fname The filename to open
 *\param num_coefs Number of coefficients in filter.
 *\return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. The worst case bit growth for the filter is returned
 * on success.
 */
int pb_load_coef_file (int *coef, char *fname, int num_coefs)
{
    char line[256];
	char *endptr;
    int i;

    double *coef_float;
    double scale = 0.0;
    double largest_tap = 0.0;

	spinerr = noerr;

    debug("pb_load_coef_file: fname=%s num_coefs=%d\n", fname, num_coefs);

	int coef_max;
	int coef_min;

    // Find the maximum and minimum values for the coefficients based
    // on the number of bits
	coef_max = 1;
	for(i=1; i < board[cur_board].fir_coef_width; i++) {
		coef_max *= 2;	
	}
	coef_min = -coef_max;
	coef_max = coef_max -1;

	debug ("pb_load_coef_file: coef_max %d, coef_min %d\n", coef_max, coef_min);


    coef_float = (double *)malloc (num_coefs * sizeof(double));

	if (!coef_float) {
		spinerr = "Out of memory: couldnt allocate internal buffer";
		debug("pb_load_coef_file: %s\n", spinerr);
		return -1;
	}

    FILE *f = fopen(fname, "r");
    if (!f) {
        spinerr = "error opening coefficent file";
		debug("pb_load_coef_file: %s\n", spinerr);
        return -1;
    }

	// Read in all the coefficients as floating point
    for (i=0; i < num_coefs; i++) {    
        if(!fgets(line, 256, f)) {
            spinerr = my_sprintf("Reached end of file before read in all coefficients (line %d)", i+1);
			debug ("pb_load_coef_file: %s\n", spinerr);
            return -1;
        }

        //coef_float[i] = atof (line);
		endptr = 0;
        coef_float[i] = strtod(line, &endptr);

		// if an error occurred converting number
		if (coef_float[i] == 0.0 && endptr== line) {
			spinerr = my_sprintf ("Error reading coefficent at line %d", i + 1);
			debug ("pb_load_coef_file: %s\n", spinerr);
			return -1;
		}

        //printf ("tap: %f\n", coef_float[i]);

		// Record the absolute value of the largest tap        
        if (coef_float[i] > largest_tap) {
            largest_tap = coef_float[i];    
        }
        if (-coef_float[i] > largest_tap) {
            largest_tap = -coef_float[i];    
        }
    }

    //printf ("largest tap %f\n", largest_tap);

    scale = (double)(coef_max+1)/largest_tap;

    //printf ("largest tap %f, scale %f\n", largest_tap, scale);

	// convert floating point coefficents to fixed point ones
    //debug ("coef_default[] = {");
    //int count = 0;

    for(i=0; i < num_coefs; i++) {

        coef[i] = (int)(coef_float[i]*scale);

        if (coef[i] > coef_max) {
            coef[i] = coef_max;    
            debug ("pb_load_coef_file: rounding coef[%d] down\n", i);
        }
        if (coef[i] < coef_min) {
            coef[i] = coef_min;    
            debug ("pb_load_coef_file: rounding coef[%d] up\n", i);
        }        

        //if (coef[i] != 0) {
        //    debug ("%d,", coef[i]);
        //    count++;
        //}

    }

    //debug("\n\nnumber taps=%d\n",count);


	free(coef_float);

	// Figure out the worst case bit growth by adding the absolute value
	// of all the coefficients together
	int sum = 0;
	for (i=0; i<num_coefs; i++) {
		if (coef[i] > 0) {
			sum += coef[i];
		}
		else {
			sum -= coef[i];
		}
	}

    return num_bits(sum);
}

/**
 * Set the parameters on the onboard FIR filter. If
 * the pb_setup_filters() function is used, filter specification is done
 * automatically and this function is not necessary.
 *
 *\param num_coefs Number of coefficients in the filter.
 *\param coef Array containing the coefficients of the filter. This array can be generated
 * from data stored in a file with the pb_load_coef_file() function. The coefficients must be even symmetric.
 *\param shift_amount Amount to shift the output of the CIC filter to the right.
 *\param dec_amount Specify by what amount the output of the FIR filter should be
 * decimated. This can be between 1 (no decimation) and 255. Take care not
 * to decimate the signal so that the resulting bandwidth is smaller than the 
 * FIR cutoff frequency, or unwanted aliasing may occur.
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_setup_fir (int num_coefs, int *coef, int shift_amount, int dec_amount)
{
    int i;
    int num_coefs_stored;

    spinerr = noerr;

    debug("pb_set_fir: num_coefs=%d shift_amount=%d dec_amount=%d\n", num_coefs, shift_amount, dec_amount);

	if (num_coefs < 0 || num_coefs > board[cur_board].fir_max_taps) {
		spinerr = "Number of coefficients out or range";
		debug("pb_set_fir: %s\n", spinerr);
		return -1;	
	}

    if (shift_amount > board[cur_board].fir_max_shift || shift_amount < 0) {
        spinerr = "shift_amount out of range";        
		debug("pb_set_fir: %s\n", spinerr);
		return -1;
    }

    if (dec_amount > board[cur_board].fir_max_decim || dec_amount < 1) {
        spinerr = "dec_amount out of range";        
		debug("pb_set_fir: %s\n", spinerr);
		return -1;
    }

    reg_write (REG_FIR_NUM_TAPS, (num_coefs-1));
    reg_write (REG_FIR_NUM_TAPS, FIR_RESET | (num_coefs-1));
    reg_write (REG_FIR_NUM_TAPS, (num_coefs-1));

    if (num_coefs % 2 == 1) {
        num_coefs_stored = num_coefs / 2 + 1;
    }
    else {
        num_coefs_stored = num_coefs / 2;        
    }

    debug ("pb_set_fir: using num_coefs_stored = %d\n", num_coefs_stored);

    // Only need to store half the coefficients, since this is even symmetric,
    // the second half of the coefficients are redundant.
    for (i = 0; i < num_coefs_stored; i++) {
        //printf ("  FIR: writing to address %d, value %d\n", i, coef[i]);

        // write the address
        reg_write (REG_FIR_COEF_ADDR, i);

        // write data
        reg_write (REG_FIR_COEF_DATA, coef[i]);

        reg_write (REG_FIR_COEF_ADDR, FIR_COEF_LOAD_EN | i);
        reg_write (REG_FIR_COEF_ADDR, i);

    }


// address 0 is the outermost coefficient, and the last address is the center
// tap. what the last address is depends on the number of taps.
// example filter: 8,-4,3,-4,8
//addr   data 
// 0  :  8
// 1  : -4
// 2  :  3


	// shift_amount and dec_amount are both 8 bit numbers
    reg_write (REG_FIR_CONTROL, (shift_amount << 8) + (dec_amount - 1));

	return 0;
}

/**
 * Retrieve the contents of the overflow registers. This can be used to find out if
 * the ADC is being driven with to large of a signal. In addition, the RadioProcessor must
 * round data values at certain points during the processing of the signal. By
 * default, this rounding is done in such a way that overflows cannot occur. However,
 * if you change the rounding procedure, this function will allow you to determine
 * if overflows have occurred. Each overflow register counts the number of overflows
 * up to 65535. If more overflows than this occur, the register will remain at 65535.
 * The overflow registers can reset by setting the reset argument of this function to 1.
 * <br><br>
 * See your manual for a detailed explanation of how the on-board rounding works.
 *
 *\param reset Set this to one to reset the overflow counters
 *\param of Pointer to a PB_OVERFLOW_STRUCT which will hold the values
 * of the overflow counter. This can be a NULL pointer if you are using
 * this function to reset the counters
 *
 *\return A negative number is returned on failure, and spinerr is set to a 
 *description of the error. 0 is returned on success.
 */
int pb_overflow(int reset, PB_OVERFLOW_STRUCT *of)
{
    int of_reg;
    int control;

    spinerr = noerr;

    if (reset) {
        control = reg_read (REG_CONTROL);
        reg_write (REG_CONTROL, control | OVERFLOW_RESET);
        reg_write (REG_CONTROL, control);        
    }
    else {
        if(of) {
            of_reg = reg_read(REG_OVERFLOW_COUNT);
            of->fir = 0x0000FFFF & (of_reg >> 16);
            of->cic = 0x0000FFFF & (of_reg);
        
            of_reg = reg_read(REG_OVERFLOW2_COUNT);

            of->average = 0x0000FFFF & (of_reg >> 16);
            of->adc =     0x0000FFFF & (of_reg);

            debug ("pb_overflow: fir: %d, cic: %d, adc: %d average: %d\n", of->fir, of->cic, of->adc, of->average);

        }
    }
       
    return 0;
}

/*
 * The onboard ADC and DAC units have several control bits which can be used
 * to control their performance characteristics. For now, users should ignore
 * this function and use the default settings. Full documentation for this
 * function will be provided soon.
 * \param adc_control
 * \param dac_control
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */

int pb_set_radio_hw(int adc_control, int dac_control)
{
    spinerr = noerr;
    
    reg_write (REG_ADC_CONTROL, adc_control);
    reg_write (REG_DAC_CONTROL, dac_control);

    return 0;
}

/**
 * The RadioProcessor contains internal registers which can be used to modify
 * the way the board works. These settings are mainly for debugging purposes
 * and not generally used during normal operation.
 * Valid bits that can be set are:
 *
 * BYPASS_MULT<br>
 * BYPASS_CIC<br>
 * BYPASS_FIR<br>
 * SELECT_AUX_DDS<br>
 * SELECT_INTERNAL_DDS<br>
 * DAC_FEEDTHROUGH<br>
 * BNC0_CLK (for boards that have selectable clock output on BNC0)<br>
 * FORCE_AVG (for boards that support averaging across separate scan calls) <br>
 *
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */

int pb_set_radio_control(unsigned int control)
{
    unsigned int old_control;
   
    spinerr = noerr;   

    //debug("pb_set_radio_control: setting control bit(s) 0x%.8x\n", control);
   
    old_control = reg_read (REG_CONTROL);
    
    reg_write (REG_CONTROL, old_control | control);    

    return 0;
}

/**
 * This function unsets bits from the control register.  Valid bits are the same ones
*  listed under pb_set_radio_control(unsigned int control).
 *
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */

int pb_unset_radio_control (unsigned int control)
{
    unsigned int old_control;

    spinerr = noerr;

    //debug("pb_unset_radio_control: un-setting control bit(s) 0x%.8x\n", control);
   
    old_control = reg_read (REG_CONTROL);

    reg_write (REG_CONTROL, old_control & (~control));

    //old_control = reg_read (REG_CONTROL);
    //debug("pb_unset_radio_control: new control word actually is 0x%.8x\n", old_control);

    return 0;
}

/*** Note: the following two functions are not used for usb devices, but I left them in rather than remove all references ***/
/**
 * Write a 32 bit (4 byte) word to the given PCI I/O address.
 * This is a low level hardware access function.
 * \param address The I/O Register to write to. This should be a multiple of 4.
 * \param data The word to write
 */
int pb_outw(unsigned int address, unsigned int data)
{
    spinerr = noerr;

    if(board[cur_board].is_usb) {
        debug("pb_outw: no support for usb devices\n");
        return -1;
    }

    // function does nothing here
    return 0;             
}

/**
 * Read a 32 bit (4 byte) word from the given PCI I/O address.
 * This is a low level hardware access function.
 * \param address The I/O Register to read. This should be a multiple of 4.
 * \return The word requested is returned.
 */
unsigned int pb_inw(unsigned int address)
{
    spinerr = noerr;

    if(board[cur_board].is_usb) {
        debug("pb_inw: no support for usb devices\n");
        return -1;
    }

    // function does nothing here
    return 0;             
}

/**
 * This function allows you to pause execution of your software for a given
 * number of milliseconds, and behaves like the sleep() function found on
 * many operating systems. This is provided because the sleep() function is not
 * standard across all operating systems/programming environments.
 * <p>This function does *NOT* interact with the hardware or pulse program at all. It
 * simply provides a portable way to access a sleep function.
 *
 * \param milliseconds Number of milliseconds to sleep (pause program) for.
 *
 */
 /* note that it was modified to use the nanosleep function in order to prevent compilation errors for osx 10.5 targets */

void pb_sleep_ms(int milliseconds)
{  
    #include <time.h>

	struct timespec req={0}, req_ret={0};  
	time_t sec=(int)(milliseconds/1000);  
	milliseconds=milliseconds-(sec*1000);  
	req.tv_sec=sec;  
	req.tv_nsec=milliseconds*1000000L;  
	nanosleep(&req,&req_ret); 

}



