/*this file modified by Tom Pratum, July 2007*/
/*derived from files originally in SpinCore spinapi dated 2-28-2007; incorporates software revision from 9/4/2007*/

#include <math.h>
#include "if.h"
#include "usb.h"
#include "util.h"
#include "caps.h"
#include "spinapi.h"

/*following needed to define Mac specific data types*/
#include <libkern/OSTypes.h>


extern char *spinerr;

extern char *noerr;

// Contains current number of instructions in pulse program
static int num_instructions=0; 

//default portbase supplied to backwards compatibility
static int port_base = 0; 

//default is pb_only with 24 output bits
static int ISA_BOARD = 0;

// FIXME: this should go in board_info
static int MAX_FLAG_BITS = 0xFFFFFF;

//2^32
double pow232 = 4294967296.0;

double last_rounded_value;

// The number of the currently selected board.
extern int cur_board;
// Number of boards present in system. 
extern int num_boards;

// This array holds the capabilties info on each board
extern BOARD_INFO board[MAX_NUM_BOARDS];

int cur_device = -1;
int cur_device_addr = 0;

//version of the DLL
char version[]="20070904";

static double shape_list[7];
static int shape_list_offset;
 
void shape_make_sin();

/**
 * This function tells the board to start programming one of the onboard 
 * devices. For all the devices, the method of programming follows the
 * following form:<br> a call to pb_start_programming(), a call to one or more
 * functions which transfer the actual data, and a call to pb_stop_programming().
 * Only one device can be programmed at a time.
 * \param device
 * Specifies which device to start programming. Valid devices are:<ul>
 * <li>PULSE_PROGRAM - The pulse program will be programmed using one of the pb_inst* instructions.
 * <li>FREQ_REGS - The frequency registers will be programmed using the pb_set_freq() function. (DDS and RadioProcessor boards only)
 * <li>TX_PHASE_REGS - The phase registers for the TX channel will be programmed using pb_set_phase() (DDS and RadioProcessor boards only)
 * <li>RX_PHASE_REGS - The phase registers for the RX channel will be programmed using pb_set_phase() (DDS enabled boards only)
 * <li>COS_PHASE_REGS - The phase registers for the cos (real) channel (RadioProcessor boards only)
 * <li>SIN_PHASE_REGS - The phase registers for the sine (imaginary) channel (RadioProcessor boards only)
 * </ul>
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_start_programming(int device)
{
    spinerr = noerr;

	int return_value;

    debug("pb_start_programming (device=%d)\n", device);

	// pb_stop_programming() didnt get called. some older code doesnt call
	// pb_stop_programming() properly, so ignore this error and continue anyway
	if (cur_device != -1) {
		debug("pb_start_programming: WARNING: pb_start_programming() called without previous stop\n", spinerr);
	}

	// Dont reset PulseBlaster if this is a new style DDS program
	if(board[cur_board].dds_prog_method == DDS_PROG_OLDPB || device == PULSE_PROGRAM) {

        debug("pb_start_programming: reset\n");

		return_value = pb_outp(port_base + 0,0); // Reset PulseBlasterDDS
		if(return_value != 0 && (!(ISA_BOARD)))
		{
            spinerr = my_strcat ("initial reset failed", spinerr);
    		debug("pb_start_programming: %s\n", spinerr);
			return return_value;
		}
	}

	if (device == PULSE_PROGRAM)
	{
		num_instructions = 0; // Clear number of instructions

		return_value = pb_outp(port_base + 2,10); // This is an instruction, therefore Bytes Per Word (BPW) = 10
		if(return_value != 0 && (!(ISA_BOARD)))
		{
            spinerr = my_strcat ("BPW=10 write failed", spinerr);
    		debug("pb_start_programming: %s\n", spinerr);
			return return_value;
		}
		
		return_value = pb_outp(port_base + 3, device); // Device = RAM
		if(return_value != 0 && (!(ISA_BOARD)))
		{
            spinerr = my_strcat ("Device=RAM write failed", spinerr);
    		debug("pb_start_programming: %s\n", spinerr);
			return return_value;
		}

		return_value = pb_outp(port_base + 4, 0); // Reset mem counter
		if(return_value != 0 && (!(ISA_BOARD)))
		{
            spinerr = my_strcat ("mem counter write failed (PULSE_PROGRAM)", spinerr);
    		debug("pb_start_programming: %s\n", spinerr);
			return return_value;
		}
		
	}
	else if (device == FREQ_REGS || device == TX_PHASE_REGS || device == RX_PHASE_REGS)
	{
		// We only need to do anything here if this the old style dds interface
		if (board[cur_board].dds_prog_method == DDS_PROG_OLDPB) {

			return_value = pb_outp(port_base + 2, 4); // This is a DDS reg, therefore BPW = 4
			if(return_value != 0 && (!(ISA_BOARD)))
			{
        	    spinerr = my_strcat ("BPW=4 write failed", spinerr);
				return return_value;
			}

			return_value = pb_outp(port_base + 3, device); // Device = FREQ_REGS, TX_PHASE_REGS, or RX_PHASE_REGS
			if(return_value != 0 && (!(ISA_BOARD)))
			{
            	spinerr = my_strcat ("Device=XXX_REGS write failed", spinerr);
				return return_value;
			}

			return_value = pb_outp(port_base + 4, 0); // Reset mem counter
			if(return_value != 0 && (!(ISA_BOARD)))
			{
        	    spinerr = my_strcat ("mem counter write failed (REGS)", spinerr);
				return return_value;
			}
		}
	}

	cur_device = device;
	cur_device_addr = 0;

	return 0;
}

/**
 * Finishes the programming for a specific onboard devices which was started by pb_start_programming(). 
 *
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_stop_programming()
{
    spinerr = noerr;

	int return_value;
	
    debug("pb_stop_programming: (device=%d)\n",cur_device);
    
	return_value = pb_outp(port_base + 7, 0);
	if(return_value != 0 && (!(ISA_BOARD)))
	{
        // spinerr will be set by pb_outp
		return return_value;
	}

	cur_device = -1;
	cur_device_addr = 0;

	return 0;
}

/**
 * Send a software trigger to the board. This will start execution of a pulse
 * program. It will also restart (trigger) a program which is currently paused due to
 * a WAIT instruction. Triggering can also be accomplished through hardware,
 * please see your board's manual for how to accomplish this.
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_start()
{
    spinerr = noerr;

	int return_value;

	return_value = pb_outp(port_base + 0, 0);
	if(return_value != 0)
	{
        spinerr = my_strcat ("+0 write failed", spinerr);
		return return_value;
	}

	return_value = pb_outp(port_base + 2, 0xFF);
	if(return_value != 0)
	{
        spinerr = my_strcat ("+2 write failed", spinerr);
		return return_value;
	}
	
	return_value = pb_outp(port_base + 3, 0xFF);
	if(return_value != 0)
	{
        spinerr = my_strcat ("+3 write failed", spinerr);
		return return_value;
	}
	
	return_value = pb_outp(port_base + 4, 0xFF);
	if(return_value != 0)
	{
        spinerr = my_strcat ("+4 write failed", spinerr);
		return return_value;
	}
	
	return_value = pb_outp(port_base + 7, 0);
	if(return_value != 0)
	{
        spinerr = my_strcat ("+7 write failed", spinerr);
		return return_value;
	}
	
	return_value = pb_outp(port_base + 1, 0);
	if(return_value != 0)
	{
        spinerr = my_strcat ("+1 write failed", spinerr);
		return return_value;
	}

	return 0;
}

/**
 * Stops output of board. Analog output will return to ground, and TTL outputs
 * will remain in the same state they were in when the stop command was received.
 * This also resets the pulseblaster so that the pulseblaster core can be run
 * again using pb_start() or a hardware trigger.
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_stop()
{
	int return_value;
    spinerr = noerr;

	return_value = pb_outp(port_base + 0,0);
	if(return_value != 0 )
	{
        // spinerr will be set by pb_outp
		return return_value;
	}

	return 0;
}

/**
 * Tell the library what clock frequency the board uses. This should be
 * called at the beginning of each program, right after you initialize the
 * board with pb_init(). Note that this does not actually set the clock
 * frequency, it simply tells the driver what frequency the board is using,
 * since this cannot (currently) be autodetected.
 * <p>
 * Also note that this frequency refers to the speed at which the PulseBlaster
 * core itself runs. On many boards, this is different than the value printed
 * on the oscillator. On RadioProcessor devices, the A/D converter and the
 * PulseBlaster core operate at the same clock frequency.
 * \param clock_freq Frequency of the clock in MHz.
 */
void pb_set_clock(double clock_freq)
{
    spinerr = noerr;

    board[cur_board].clock = clock_freq / 1000.0; // Put in GHz (for ns timescale)
}

/**
 * Write the given frequency to a frequency register on a DDS enabled board. To do this, first call
 * pb_start_programming(), and pass it FREQ_REGS. The first call pb_set_freq() will then program
 * frequency register 0, the second call will program frequency register 1, etc.
 * When you have programmed all the registers you intend to, call pb_stop_programming()
 *
 * \param freq The frequency in MHz to be programmed to the register.
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 *
 */
int pb_set_freq(double freq)
{
    spinerr = noerr;

	unsigned int freq_byte = 0xFF000000;//Bit Mask to send 1 byte at a time
	double dds_clock;
	UInt32 freq_word;
    UInt32 freq_word2;
	unsigned int temp_byte;
	int return_value;
	int i = 0;

    dds_clock = board[cur_board].clock;
    freq_word = (UInt32) lrint(((freq * pow232) / (dds_clock*1000.0))); // Desired data to transfer
    freq_word2 = (UInt32) lrint(((freq * pow232) / (dds_clock*1000.0*board[cur_board].dds_clock_mult)));

	debug("pb_set_freq: address:%d freq:%9.3f freq_word:%lu clock:%9.5f\n", cur_device_addr, freq, freq_word, dds_clock);

	// Check if use has already written to all registers
	if (cur_device_addr >= board[cur_board].num_freq0) {
		spinerr = "Frequency registers full";
		debug ("pb_set_freq: %s\n", spinerr);
		return -1;
	}

	if (board[cur_board].dds_prog_method == DDS_PROG_OLDPB) {
        debug("pb_set_freq: using old programming method\n");

		for(i = 0; i<4; i++) // Loop to send 4 bytes
		{
			temp_byte = freq_byte & freq_word;  // Get current byte to transfer
			temp_byte >>= 24; // Shift data into LSB
			return_value = pb_outp(port_base + 6, temp_byte); // Send byte to PulseBlasterDDS
			if(return_value != 0 && (!(ISA_BOARD)))
			{
				return -1;
			}

			freq_word <<= 8; // Shift word up one byte so that next byte to transfer is in MSB
		}
	}
	else {
        debug("pb_set_freq: using new programming method\n");
		dds_freq_extreg (cur_board, cur_device_addr, freq_word, freq_word2);
	}

	cur_device_addr++;

	return 0;
}

/**
 * Write the given phase to a phase register on DDS enabled boards. To do this, first call
 * pb_start_programming(), and specify the appropriate bank of phase
 * registers (such as TX_PHASE, RX_PHASE, etc) as the argument. The first call pb_set_phase() will then program
 * phase register 0, the second call will program phase register 1, etc.
 * When you have programmed all the registers you intend to, call pb_stop_programming()
 * <br>
 * The given phase value may be rounded to fit the precision of the board.
 *
 * \param phase The phase in degrees to be programmed to the register.
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 *
 */
int pb_set_phase(double phase)
{
    spinerr = noerr;
	
	unsigned int phase_byte = 0xFF000000;//Bit Mask to send 1 byte at a time

	double temp = phase/360.0;

    // fix phase to be 0 <phase <360.
    /* the rather odd code that was originally here is below:
    while (temp >= 1.0) {
        temp -= 1.0;
    }
    while (temp < 0.0) {
        temp += 1.0;
    }*/

	temp = temp - trunc(temp);
	double temp2 = temp *pow232;
	/*it was necessary here to explicitly state that the integer is 32 bits*/
	UInt32 phase_word = (UInt32)llrint(temp2);

	debug("pb_set_phase: phase %9.2f, phase word %lu, device address %lu\n",phase,phase_word,cur_device_addr);
	unsigned int temp_byte;
	int return_value;
	int i = 0;


    int max_phase_regs = 0;
    
    if(cur_device == COS_PHASE_REGS) {
        max_phase_regs = board[cur_board].num_phase0;
    }
    if(cur_device == SIN_PHASE_REGS) {
        max_phase_regs = board[cur_board].num_phase1;
    }
    if(cur_device == TX_PHASE_REGS) {
        if(board[cur_board].is_radioprocessor) {
            max_phase_regs = board[cur_board].num_phase2;            
        }
        else {
            max_phase_regs = board[cur_board].num_phase1;
        }
    }
    if(cur_device == RX_PHASE_REGS) {
        max_phase_regs = board[cur_board].num_phase0;
    }

	// Check if use has already written to all registers
	if (cur_device_addr >= max_phase_regs) {
		spinerr = "Phase registers full";
		debug ("pb_set_phase: %s\n", spinerr);
		return -1;
	}

	if (board[cur_board].dds_prog_method == DDS_PROG_OLDPB) {
		for(i = 0; i<4; i++) // Loop to send 4 bytes
		{
			temp_byte = phase_byte & phase_word; // Get current byte to transfer
			temp_byte >>= 24; // Shift data into LSB
			return_value = pb_outp(port_base + 6, temp_byte); // Send byte to PulseBlasterDDS
			if(return_value != 0 && (!(ISA_BOARD))) // conio's _outp returns value passed to it
			{
				return -1;
			}

			phase_word <<= 8; // Shift word up one byte so that next byte to transfer is in MSB
		}
	}
	else {
		dds_phase_extreg (cur_board, cur_device, cur_device_addr, phase_word);
	}

	cur_device_addr++;

	return 0;
}



/**
 * \internal
 * Program a dds register using the extended register method. (this is how to control the dds if
 * dds_prog_method=DDS_PROG_EXTREG)
 * \param addr which register to program
 * \param cur_board which board to program
 * \param freq_word Frequency word for the "normal" speed DDS outputs (sin, cos, internal)
 * \param freq_word2 Frequency word for the "Fast" speed DDS, (DDS that feeds the DAC)
 */
int dds_freq_extreg (int cur_board, int addr, UInt32 freq_word, UInt32 freq_word2)
{
  UInt32 control_word;

  reg_write (REG_DDS_DATA, (unsigned int)freq_word);
  // REG_DDS_DATA2 writes the frequency word to the DDS unit which drives the DAC. In some
  // designs, that DDS runs at a speed FASTER than the other DDS units. Thus
  // it needs a seperate frequency word to ensure that the frequency it outputs
  // is the same as the slower DDS. See board.dds_clock_mult for how much faster
  // it runs
  reg_write (REG_DDS_DATA2, (unsigned int)freq_word2);

  // put the address on the registers
  control_word = DDS_WRITE_SEL | ((0x0FF & addr) << 14);
  reg_write (REG_DDS_CONTROL, (unsigned int)control_word);
  // do the actual write
  reg_write (REG_DDS_CONTROL, (unsigned int)(control_word | DDS_FREQ_WE));
  reg_write (REG_DDS_CONTROL, (unsigned int)control_word);
  // reset the control 
  reg_write (REG_DDS_CONTROL, (unsigned int)DDS_RUN); // in modern designs, DDS_RUN is not connected to anything and the DDS cores run all the time, but need this here for compatibilities sake

  return 0;
}

/**
 *\internal
 *\param cur_board which board to program
 *\param phase_bank
 *\param addr which register to program
 */ 
int dds_phase_extreg (int cur_board, int phase_bank, int addr, UInt32 phase_word)
{
  UInt32 control_word = DDS_WRITE_SEL;
  UInt32 we = 0;

  switch (phase_bank) {

    case SIN_PHASE_REGS:
      control_word |= (addr << 10);
      we = DDS_TX_PHASE_WE; // the sin dds output is called tx internally
      break;
    case COS_PHASE_REGS:
      control_word |= (addr << 6);
      we = DDS_RX_PHASE_WE; // the cos dds output is called rx internally
      break;
    case TX_PHASE_REGS:
      /*printf("DDS_WRITE_SEL %lu, control_word %lu\n", DDS_WRITE_SEL,control_word);*/
      control_word |= (addr << 19);
      /*printf("DDS_WRITE_SEL %lu, control_word %lu\n", DDS_WRITE_SEL,control_word);*/
      we = DDS_REF_PHASE_WE; // the tx dds output is called ref internally
      break;

    default:
    
        debug("dds_phase_extreg: invalid phase bank\n");
        break;

  }

  reg_write (REG_DDS_DATA, (unsigned int)phase_word);
  reg_write (REG_DDS_DATA2, (unsigned int)phase_word); // "Fast" DDS uses same phase info as other DDSes

  reg_write (REG_DDS_CONTROL, (unsigned int)control_word);
  reg_write (REG_DDS_CONTROL, (unsigned int)(control_word | we));    
  reg_write (REG_DDS_CONTROL, (unsigned int)control_word);
  reg_write (REG_DDS_CONTROL, (unsigned int)DDS_RUN);

  return 0;
}

char* pb_get_version()
{
    spinerr = noerr;

	return version;
}

/**
 * Get the firmware version on the board. This is not supported on all boards.
 *
 *\return Returns the firmware id as described above. A 0 is returned if the
 *firmware id feature is not available on this version of the board.
 */
int pb_get_firmware_id()
{
    spinerr = noerr;

	if(board[cur_board].has_firmware_id) {

        debug("pb_get_firmware_id: has id\n");
        if(board[cur_board].has_firmware_id == 1) {
    		return reg_read(REG_FIRMWARE_ID);	
        }
        else {
            return reg_read(0x15); // usb boards have the firmware ID stored in register 0x15
        }
	}
	else {
		return 0;	
	}
}

/**
 * Return the most recent error string. Anytime a function (such as pb_init(),
 * pb_start_programming(), etc.) encounters an error, this function will return
 * a description of what went wrong.
 *
 * \return A string describing the last error is returned. A string containing
 * "No Error" is returned if the last function call was successfull.
 */
char* pb_get_error()
{
	return spinerr;	
}

/**
 * This function sets the RadioProcessor to its default state. It has no effect
 * on any other SpinCore product. This function should generally be called after pb_init()
 * to make sure the RadioProcessor is in a usable state. It is REQUIRED that this
 * be called at least once after the board is powered on. However, there are a few
 * circumstances when you would not call this function. In the case where you had
 * one program that configured the RadioProcessor, and another seperate program
 * which simply called pb_start() to start the experiment, you would NOT call
 * pb_set_defaults() in the second program because this would overwrite the
 * configuration set by the first program.
 *
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */

int pb_set_defaults()
{
    spinerr = noerr;
    int i;
    

    if(board[cur_board].is_radioprocessor) {
        debug("pb_set_defaults: setting defaults for RadioProcessor\n");

        int adc_control = 3;
        int dac_control = 0;
        pb_set_radio_hw(adc_control, dac_control);
        
        reg_write(REG_CONTROL, 0);

        if(board[cur_board].supports_dds_shape) {
            debug("pb_set_defaults: setting shape defaults\n");
            for(i=0; i < board[cur_board].num_shape; i++) {
                shape_list[i] = 0.0;
            }
            shape_list_offset = 0;

            // Make sure DDS unit has a sinewave loaded, and amp register 0 has a value of 1.0
            // This will ensure the board works properly with shape-unaware code
            float dds_data[1024];
            shape_make_sin(dds_data);
            pb_dds_load(dds_data, DEVICE_DDS);

            pb_set_amp(1.0, 0);
        }

        return 0;
    }
    else {
        spinerr = "This function is for RadioProcessor only and does nothing on your board";
        debug("pb_set_defaults: %s\n", spinerr);        
        return -1;
    }
   
}

/**
 * Program an instruction of the pulse program. 
 *
 *\param freq Selects which frequency register to use
 *\param cos_phase Selects which phase register to use for the cos (real) channel
 *\param sin_phase Selects which phase register to use for the sin (imaginary) channel
 *\param tx_phase Selects which phase register to use for the TX channel
 *\param tx_enable When this is 1, the TX channel will be output on the Analog
 * Out connector. When this is 0, Analog Out channel will be turned off.
 *\param phase_reset When this is 1, the phase of all DDS channels will be reset
 * to their time=0 phase. They will stay in this state until the value of this
 * bit returns to 0.
 *\param trigger_scan When this is 1, a scan will be triggered. To start a second
 * scan, this bit must be set to 0 and then back to 1.
 *\param flags Controls the state of the user available digital out pins. Since
 * there are 6 user outputs, only the lower 6 bits of this parameter are used.
 * Bits 1 and 0 control BNC1 and BNC0 respectively.
 *\param inst Which instruction to use. See manual for details.
 *\param inst_data Some instructions require additional data. This allows that data
 * to be specified. See manual for details.
 *\param length Time until the next instruction is executed in nanoseconds
 */
int pb_inst_radio(int freq, int cos_phase, int sin_phase, int tx_phase, int tx_enable, int phase_reset, int trigger_scan, int flags, int inst, int inst_data, double length)
{
    spinerr = noerr;
    int flag_word = 0;
    
    if(freq >= board[cur_board].num_freq0 || freq < 0) {
        spinerr = "Frequency register out of range";
        debug("pb_inst_radio: %s\n", spinerr);
        return -1;
    }

    if(cos_phase >= board[cur_board].num_phase0 || cos_phase < 0) {
        spinerr = "Cos phase register out of range";
        debug("pb_inst_radio: %s\n", spinerr);
        return -1;
    }

    if(sin_phase >= board[cur_board].num_phase1 || sin_phase < 0) {
        spinerr = "Sin phase register out of range";
        debug("pb_inst_radio: %s\n", spinerr);
        return -1;
    }

    if(tx_phase >= board[cur_board].num_phase2 || tx_phase < 0) {
        spinerr = "TX phase register out of range";
        debug("pb_inst_radio: %s\n", spinerr);
        return -1;
    }

    if(board[cur_board].supports_dds_shape) {
        debug("pb_inst_radio: passing on values to pb_inst_radio_shape\n");
        int amp = 0;
        // Note: this will only work correctly if a value of 1.0 is written into amplitude register 0 (which is the default, and is taken care of by pb_set_defaults())
        return pb_inst_radio_shape(freq, cos_phase, sin_phase, tx_phase, tx_enable, phase_reset, trigger_scan, 0, amp, flags, inst, inst_data, length);
    }
	
	if(board[cur_board].acquisition_disabled == 1 && trigger_scan == 1) {
        spinerr = "Your version of the RadioProcessor does not support data acquisition.";
        debug("pb_inst_radio: %s\n", spinerr);
        return -1;
    }
	
    flag_word |= (0x03 & sin_phase)    << 22; // 2 bits
    flag_word |= (0x03 & cos_phase)    << 20; // 2 bits
    flag_word |= (0x7F & tx_phase)     << 13; // 7 bits
    flag_word |= (0x01 & tx_enable)    << 12; // 1 bit
    flag_word |= (0x0F & freq)         << 8;  // 4 bits
    flag_word |= (0x01 & trigger_scan) << 7;  // 1 bit
    flag_word |= (0x01 & phase_reset)  << 6;  // 1 bit

    flag_word |= (0x3F & flags);
    
    debug("pb_inst_radio: using standard RadioProcessor flag_word partitioning scheme(no shape support)\n",inst,inst_data,length);
    debug("pb_inst_radio: inst=%d, inst_data=%d,length=%f\n",inst,inst_data,length);
    
    return pb_inst_pbonly (flag_word, inst, inst_data, length);
}

/**
 * Write an instruction that makes use of the pulse shape feature of some RadioProcessor boards. This adds two new paramters, use_shape and amp, which control the shape feature. All other parameters are identical to the pb_inst_radio() function. If you do not wish to use the shape feature, the pb_inst_radio() function can be used instead.
 * \param use_shape Select whether or not to use shaped pulses. If this is 0, a regular non-shaped pulse (hard pulse) is output. If it is nonzero, the shaped pulse is used. The pulse shape waveform can be set using the pb_dds_load() function.
 * \param amp Select which amplitude register to use. The values of the amplitude registers can be set with pb_set_amp()
 */
int pb_inst_radio_shape(int freq, int cos_phase, int sin_phase, int tx_phase, int tx_enable, int phase_reset, int trigger_scan, int use_shape, int amp, int flags, int inst, int inst_data, double length)
{
    spinerr = noerr;
    unsigned int flag_word = 0;

    int shape_period = -1;
    int i;

    debug("pb_inst_radio_shape: freq=%d, cos_phase=%d, sin_phase=%d, tx_phase=%d, ", freq, cos_phase, sin_phase, tx_phase);
    debug("tx_enable=%d, phase_reset=%d, trigger_scan=%d, use_shape=%d,", tx_enable, phase_reset, trigger_scan, use_shape);
    debug(" amp=%d\n", amp);
    debug("pb_inst_radio_shape: flag=0x%x, inst=%d, inst_data=%d, length=%e\n", flags, inst, inst_data, length);

    if(!board[cur_board].supports_dds_shape) {
        spinerr = "Board does not support DDS shape capabilities";
        debug("pb_inst_radio_shape: %s\n", spinerr);
        return -1;
    }

	if(board[cur_board].acquisition_disabled == 1 && trigger_scan == 1) {
		spinerr = "Your version of the RadioProcessor does not support data acquisition.";
        debug("pb_inst_radio: %s\n", spinerr);
        return -1;
    }
	
	
    if(amp >= board[cur_board].num_amp || amp < 0) {
        spinerr = "Amplitude register out or range";
        debug("pb_inst_radio_shape: %s\n", spinerr);
        return -1;
    }

    if(use_shape) {
        // check if this shape period has already been programmed
        for(i=0; i < shape_list_offset; i++) {
            if(shape_list[i] == length) {
                shape_period = i;
                debug("pb_inst_radio_shape: using shape register %d\n", shape_period);
                break;
            }
        }
        // if not, program it
        if(shape_period == -1) {
            // if there is room for another shape period
            if(shape_list_offset < board[cur_board].num_shape) {
                debug("pb_inst_radio_shape: adding shape period %f to register %d\n", length, shape_list_offset);
                set_shape_period(length, shape_list_offset);
                shape_period = shape_list_offset;
                shape_list[shape_list_offset] = length;
                shape_list_offset++;
            }
            else {
                spinerr = "No more shape period registers available";
                debug("pb_inst_radio_shape: %s\n", spinerr);
                return -1;
            }
        }

    }
    else {
        // when shape_period is the maximum value + 1, this indicates that we are not using a shape at all
        // for example, if there are 7 shape period registers
        // shape_period=0-6 are some period
        // and shape_period=7 indicates no shape at all
        // num_shape must always fit the form 2^x-1
        debug("pb_inst_radio_shape: bypassing shape\n");
        shape_period = board[cur_board].num_shape;
    }

    // this is the r2 flag word partition
	if(board[cur_board].custom_design == 1)
    {	//firmware revision 10-10, in which some register select lines have been converted to TTL output bits and re-arranged
		flag_word |= (0x03 & sin_phase)    << 22; // 2 bits
		flag_word |= (0x03 & cos_phase)    << 20; // 2 bits
		flag_word |= (0x01 & tx_phase)     << 16; // 1 bit
		flag_word |= (0x01 & tx_enable)    << 15; // 1 bit
		flag_word |= (0x03 & freq)         << 11; // 2 bits
		flag_word |= (0x01 & trigger_scan) << 10; // 1 bit
		flag_word |= (0x01 & phase_reset)  << 9;  // 1 bit
		flag_word |= (0x07 & shape_period) << 6;  // 3 bits
		flag_word |= (0x03 & amp)          << 4;  // 2 bits
		flag_word |= (0x1C0 & flags)       << 11; //upper 3 bits
		flag_word |= (0x30 & flags)		   << 9;  //middle 2 bits
		flag_word |= (0x0F & flags);              // bottom 4 bits
		
		debug("pb_inst_radio_shape: using TOPSPIN flag_word partitioning scheme\n");
    }
	else if (board[cur_board].custom_design == 2)
    {   //firmware revision 12-6, in which some register select lines have been converted to TTL output bits and re-arranged
	    flag_word |= (0x03 & sin_phase)    << 22; // 2 bits
		flag_word |= (0x03 & cos_phase)    << 20; // 2 bits
		flag_word |= (0x03 & tx_phase)     << 18; // 2 bits
		flag_word |= (0x01 & tx_enable)    << 15; // 1 bit
		flag_word |= (0x03 & freq)         << 16; // 2 bits
		flag_word |= (0x01 & trigger_scan) << 8; // 1 bit
		flag_word |= (0x01 & phase_reset)  << 9;  // 1 bit
		flag_word |= (0x07 & shape_period) << 12;  // 3 bits
		flag_word |= (0x03 & amp)          << 10;  // 2 bits
		flag_word |= (0xFF & flags);              // 8 bits
        
        debug("pb_inst_radio_shape: using PROGRESSION flag_word partitioning scheme\n");    
	}
	else if(board[cur_board].custom_design == 3)
	{   //firmware revision 10-15, PBDDS-i-300. RP board in which aquisition has been disabled and there are 9 TTL outputs
	    unsigned int flags8to5 = (flags & 0x1E0) >> 5;
	    unsigned int flag4     = (flags & 0x010) >> 4;
	    unsigned int flags3to0 = (flags & 0x00F);
	
        flag_word |= (0x0F & flags8to5)    << 20;  // 4 bits 
        flag_word |= (0x0F & tx_phase)     << 16;  // 4 bits
		flag_word |= (0x01 & tx_enable)    << 15;  // 1 bit
		flag_word |= (0x0F & freq)         << 11;  // 4 bits
		flag_word |= (0x01 & flag4)        << 10;  // 1 bit
		flag_word |= (0x01 & phase_reset)  << 9;   // 1 bit
		flag_word |= (0x07 & shape_period) << 6;   // 3 bits
		flag_word |= (0x03 & amp)          << 4;   // 2 bits
		flag_word |= (0x0F & flags3to0);           // 4 bits
        
        debug("pb_inst_radio_shape: using PBDDS-300 flag_word partitioning scheme\n");    
    }
	else{
        //this is the standard r2 flag word partition
		flag_word |= (0x03 & sin_phase)    << 22; // 2 bits
		flag_word |= (0x03 & cos_phase)    << 20; // 2 bits
		flag_word |= (0x0F & tx_phase)     << 16; // 4 bits
		flag_word |= (0x01 & tx_enable)    << 15; // 1 bit
		flag_word |= (0x0F & freq)         << 11; // 4 bits
		flag_word |= (0x01 & trigger_scan) << 10; // 1 bit
		flag_word |= (0x01 & phase_reset)  << 9;  // 1 bit
		flag_word |= (0x07 & shape_period) << 6;  // 3 bits
		flag_word |= (0x03 & amp)          << 4;  // 2 bits
		flag_word |= (0x0F & flags);              // 4 bits
		
		debug("pb_inst_radio_shape: using standard flag_word partitioning scheme\n");
		if(board[cur_board].custom_design != 0)
            debug("WARNING 002: You are using the wrong instruction(i.e. pb_inst_onerf, pb_inst_radio, etc), Please refer to your manual.\nContact www.SpinCore.com for help.\n");
    }
	
    debug("pb_inst_radio_shape: shape :%d, amp: %d\n", shape_period, amp);
    
    return pb_inst_pbonly (flag_word, inst, inst_data, length);
}

/**
 * Program an instruction of the pulse program. For use with PBDDS-I-300.
 *
 *\param freq Selects which frequency register to use
 *\param tx_phase Selects which phase register to use for the TX channel
 *\param tx_enable When this is 1, the TX channel will be output on the Analog
 * Out connector. When this is 0, Analog Out channel will be turned off.
 *\param phase_reset When this is 1, the phase of all DDS channels will be reset
 * to their time=0 phase. They will stay in this state until the value of this
 * bit returns to 0.
 *\param flags Controls the state of the user available digital out pins. Since
 * there are 9 user outputs, only the lower 9 bits of this parameter are used.
 * Bits 1 and 0 control BNC1 and BNC0 respectively.
 *\param inst Which instruction to use. See manual for details.
 *\param inst_data Some instructions require additional data. This allows that data
 * to be specified. See manual for details.
 *\param length Time until the next instruction is executed in nanoseconds
 */
int pb_inst_dds(int freq, int tx_phase, int tx_enable, int phase_reset, int flags, int inst, int inst_data, double length)
{
    debug("pb_inst_dds: passing on values to pb_inst_radio\n");
    return pb_inst_radio(freq, 0, 0, tx_phase, tx_enable, phase_reset, 0, flags, inst, inst_data, length);
}

/**
 * Write an instruction that makes use of the pulse shape feature of the PulseBlasterDDS-I-300 board. 
 * This requires two new paramters, use_shape and amp, which control the shape feature. All other 
 * parameters are identical to the pb_inst_dds() function. If you do not wish to use the shape feature, 
 * the pb_inst_dds() function can be used instead.
 * 
 * \param use_shape Select whether or not to use shaped pulses. If this is 0, a regular non-shaped pulse 
 * (hard pulse) is output. If it is nonzero, the shaped pulse is used. The pulse shape waveform can be set 
 * using the pb_dds_load() function.
 * \param amp Select which amplitude register to use. The values of the amplitude registers can be set with pb_set_amp()
 */
int pb_inst_dds_shape(int freq, int tx_phase, int tx_enable, int phase_reset, int use_shape, int amp, int flags, int inst, int inst_data, double length)
{
    debug("pb_inst_dds_shape: passing on values to pb_inst_radio_shape\n");
    return pb_inst_radio_shape(freq, 0, 0, tx_phase, tx_enable, phase_reset, 0, use_shape, amp, flags, inst, inst_data, length);
}

/**
 * \internal
 * This function generates one period of a sine wave, and stores it in the given array.
 * \param shape_data array of 1024 floats, which will form the array
 */
void shape_make_sin(float *shape_data)
{
    int i;

    double pi = 3.1415926;

    for(i=0; i < 1024; i++) {
        shape_data[i] = sin(2.0*pi*((float)i/1024.0));
    }    
}
/**
 * \internal
 * This function sets the given shape period register with the given value. It
 * is used by pb_inst_radio_shape() to manage the values of the shape registers.
 * For now, we hide the existance of these registers from the user.
 * \param period in ns
 * \param addr Address of shape period register to write to
 */
int set_shape_period(double period, int addr)
{
    spinerr = noerr;

    if(!board[cur_board].supports_dds_shape) {
        spinerr = "DDS Shape capabilities not supported on this board";
        debug("pb_set_shape_period: %s\n", spinerr);
        return -1;
    }

    // convert period in ns to frequency in MHz
    double freq = (1.0/period)*1e3;

    if(addr >= board[cur_board].num_shape) {
        spinerr ="Shape period registers full";
        debug("pb_set_shape_period: %s\n", spinerr);
        return -1;
    }

    debug("pb_set_shape_period: addr=%d, period=%f, freq=%f\n", addr, period, freq);

    double dds_clock;
    double dds_clock_mult;
    unsigned int freq_word2;

    dds_clock = board[cur_board].clock;
    dds_clock_mult = board[cur_board].dds_clock_mult;
    freq_word2 = (unsigned int) rint(((freq * pow232) / (dds_clock*1000.0*dds_clock_mult)));

    //debug("pb_set_shape_period: freq_word2: 0x%x\n", freq_word2);

    reg_write (REG_DDS_DATA2, freq_word2);

    reg_write (REG_SHAPE_CONTROL, SHAPE_WRITE_ADDR_SEL | (0x07 & addr));
    reg_write (REG_SHAPE_CONTROL, SHAPE_WRITE_ADDR_SEL | SHAPE_FREQ_WE | (0x07 & addr));
    reg_write (REG_SHAPE_CONTROL, SHAPE_WRITE_ADDR_SEL | (0x07 & addr));
    reg_write (REG_SHAPE_CONTROL, 0);

    return 0;
}

/**
 * Set the value of one of the amplitude registers.
 * \param amp Amplitude value. 0.0-1.0
 * \param addr Address of register to write to
 * 
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_set_amp(float amp, int addr)
{
    spinerr = noerr;

    int freq_word2;

    if(!board[cur_board].supports_dds_shape) {
        spinerr = "DDS Shape capabilities not supported on this board";
        debug("pb_set_shape_period: %s\n", spinerr);
        return -1;
    }

    if(addr >= board[cur_board].num_amp) {
        spinerr ="Amplitude registers full";
        debug("pb_set_amp: %s\n", spinerr);
        return -1;
    }

    debug("pb_set_amp: addr=%d, amp=%f\n", addr, amp);

    // FIXME: it may be possible to have negative amplitudes as well, though this could also
    // be accomplished through the phase registers
    if(amp > 1.0 || amp < 0.0) {
        spinerr = "Amplitude must be between 0.0 and 1.0, inclusive";
        debug("pb_set_amp: %s\n", spinerr);
        return -1;
    }


    freq_word2 = (int)(amp*(float)(1 << 14));
    //debug("pb_set_amp: freq_word2 is 0x%x\n", freq_word2);

    reg_write (REG_DDS_DATA2, (unsigned int)freq_word2);

    unsigned int control = SHAPE_WRITE_ADDR_SEL | (0x03 & addr);

    reg_write (REG_SHAPE_CONTROL, control);

    // FIXME: I think address select is no longer relevant
    pb_set_radio_control(0x10); // the imag_add is the address select
    pb_set_radio_control(0x08); // the real_add signal is the we
    pb_unset_radio_control(0x08);
    pb_unset_radio_control(0x10);
    reg_write (REG_SHAPE_CONTROL, control);


    reg_write (REG_DDS_DATA2, (unsigned int)0);

    return 0;
}


/**
 * Load the DDS with the given waveform. There are two different waveforms that can be
 * loaded. 
 * <ul>
 * <li>DEVICE_DDS - This is for the DDS module itself. By default, it is loaded with a sine wave,
 * and if you don't wish to change that or use shaped pulses, you do not need to use this function.
 * Otherwise this waveform can be loaded with any arbitrary waveform that will be used instead of a sine
 * wave.
 * <li>DEVICE_SHAPE - This waveform is for the shape function. This controls the shape used,
 * if you enable the use_shape parameters of pb_inst_radio_shape(). For example, if you wish to
 * use soft pulses, this could be loaded with the values for the sinc function.
 * </ul>
 * \param data This should be an array of 1024 floats that represent a single period of the waveform
 * you want to have loaded. The range for each data point is from -1.0 to 1.0 
 * \param device Device you wish to program the waveform to. Can be DEVICE_SHAPE or DEVICE_DDS
 * \return A negative number is returned on failure, and spinerr is set to a 
 * description of the error. 0 is returned on success.
 */
int pb_dds_load(float *data, int device)
{
    spinerr = noerr;

    if(!board[cur_board].supports_dds_shape) {
        spinerr = "DDS Shape capabilities not supported on this board";
        debug("pb_dds_load: %s\n", spinerr);
        return -1;
    }

    int i;
    int val;

    char raw_data[2048];

    debug("pb_dds_load: writing to device 0x%x\n", device);

    for(i = 0; i < 2048; i+=2) {
        if(data[i/2] > 1.0 || data[i/2] < -1.0) {
            spinerr = "Data must be between -1.0 and 1.0, inclusive";
            debug("pb_dds_load: %s (point %d)\n", spinerr, i/2);
            return -1;
        }

        switch(device) {
            case DEVICE_SHAPE:
                val = -(int)(data[i/2]*(float)(1<<14));
                if(val > 16383) {
                    val = 16383;
                }
                if(val < -16384) {
                    val = -16384;
                }
                break;
            case DEVICE_DDS:
                val = -(int)(data[i/2]*8192.0);   
                if(val > 8191) {
                    val = 8191;
                }
                if(val < -8191) {
                    val = -8191;
                }

                break;
            default:
                spinerr="Invalid device number\n";
                debug("pb_dds_load: %s\n");
                return -1;
        }

        raw_data[i]  = 0x0FF & (val);
        raw_data[i+1]= 0x0FF & (val >> 8);

    }

    switch(device) {
        case DEVICE_SHAPE:
            reg_write(REG_SHAPE_CONTROL, SHAPE_SHAPERAM_WRITE_SEL);
            break;
        case DEVICE_DDS:
            reg_write(REG_SHAPE_CONTROL, SHAPE_DDSRAM_WRITE_SEL);        
            break;
        default:
            spinerr="Invalid device number\n";
            debug("pb_dds_load: %s\n");
            return -1;
    }
      

    ram_write(BANK_DDSRAM, 0, 2048, raw_data);

    reg_write(REG_SHAPE_CONTROL, 0);

    return 0;
}


/**
 * Program an instruction of a pulse program to the board. This function
 * programs instructions for boards with 2 DDS outputs. (such as PulseBlasterDDS-III)
 * There are other forms
 * of this instruction (see below) design to program boards with differing
 * instruction formats.<br>
 * 
 * Instructions can also be programmed by calling pb_inst(), but you must
 * use an appropriate #define before including spinapi.h. See the example programs
 * for how to do this.
 *
 * \param freq The frequency register to use for this instruction
 * \param tx_phase The TX phase register to use for this instruction
 * \param tx_output_enable Set to ANALOG_ON to enable output, or ANALOG_OFF to output nothing
 * \param rx_phase The RX phase register to use for this instruction
 * \param rx_output_enable Set to ANALOG_ON to enable output, or ANALOG_OFF to output nothing
 * \param flags Set every bit to one for each flag you want to set high
 *
 * \param inst Specify the instruction you want. Valid instructions are:
 *<table border="0">
 *<tr><td>Opcode #</td><td>Instruction</td><td>Meaning of inst_data field</td></tr>
 *<tr><td>0</td><td>CONTINUE</td><td>Not Used</td></tr>
 *<tr><td>1</td><td>STOP</td><td>Not Used</td></tr>
 *<tr><td>2</td><td>LOOP</td><td>Number of desired loops</td></tr>
 *<tr><td>3</td><td>END_LOOP</td><td>Address of instruction originating loop</td></tr>
 *<tr><td>4</td><td>JSR</td><td>Address of first instruction in subroutine</td></tr>
 *<tr><td>5</td><td>RTS</td><td>Not Used</td></tr>
 *<tr><td>6</td><td>BRANCH</td><td>Address of instruction to branch to</td></tr>
 *<tr><td>7</td><td>LONG_DELAY</td><td>Number of desired repetitions</td></tr>
 *<tr><td>8</td><td>WAIT</td><td>Not Used</td></tr>
 *</table>
 * See your board manual for a detailed description of each instruction.
 *
 *
 * \param inst_data Instruction specific data. Internally this is a 20 bit unsigned number, so the largest value that can be passed is 2^20-1 (the largest value possible for a 20 bit number). See above table to find out what this means for each instruction.
 * \param length Length of this instruction in nanoseconds.
 * \return The address of the created instruction is returned. This can be used
 * as the branch address for any branch instructions. A negative number is
 * returned on failure, and spinerr is set to a description of the error.
 */
int pb_inst_tworf(int freq, int tx_phase, int tx_output_enable, int rx_phase, int rx_output_enable, int flags, int inst, int inst_data, double length)
{
    spinerr = noerr;

	unsigned int OCW;

    // 
    // FIXME: this should check against the values on the BOARD_INFO structure instead
    // of predefined values
    //

	//Check for valid passed parameters
	if(freq < 0 || freq > MAX_FREQ_REGS)
	{
        spinerr = "Frequency register out of range";
        debug("pb_inst_tworf: %s\n");
		return -99;
	}

	if(tx_phase < 0 || tx_phase > MAX_FREQ_REGS)
	{
        spinerr = "TX phase register out of range";
        debug("pb_inst_tworf: %s\n");
		return -98;
	}
	if(tx_output_enable != ANALOG_OFF)
	{
        tx_output_enable = ANALOG_ON;
	}
	if(rx_phase < 0 || rx_phase > MAX_PHASE_REGS)
	{
        spinerr = "RX phase register out of range";
        debug("pb_inst_tworf: %s\n");
		return -96;
	}
	if(rx_output_enable != ANALOG_OFF)
	{
        rx_output_enable = ANALOG_ON;
	}
	if(flags < 0 || flags > MAX_FLAG_BITS)
	{
        spinerr = "Too many flag bits";
        debug("pb_inst_tworf: %s\n");
		return -94;
	}
	
	OCW = (freq << 20) | (tx_phase << 16) | (!(tx_output_enable) << 11) | (rx_phase << 12) | (!(rx_output_enable) << 10) | (flags & MAX_FLAG_BITS);

	return pb_inst_pbonly(OCW, inst, inst_data, length);
}

/**
 * This is the instruction programming function for boards with only one DDS
 * output channel. 
 * Syntax is identical to that of
 * pb_inst_tworf(), but the second RF channel is not used.
 */
int pb_inst_onerf(int freq, int phase, int rf_output_enable, int flags, int inst, int inst_data, double length)
{
    spinerr = noerr;

	return pb_inst_tworf(freq, phase, rf_output_enable, 0, 1, flags, inst, inst_data, length);
}

/**
 * This is the instruction programming function for boards without a DDS. 
 * (for example PulseBlaster and PulseBlasterESR boards). Syntax is identical to that of
 * pb_inst_tworf(), except that the parameters pertaining to the analog outputs
 * are not used.
 */
int pb_inst_pbonly(int flags, int inst, int inst_data, double length)
{
    int delay;
    double pb_clock, clock_period;

    spinerr = noerr;

    pb_clock = board[cur_board].clock * board[cur_board].pb_clock_mult;
	clock_period = 1.0 / pb_clock;

	delay = (int) rint((length * pb_clock) - 3.0); //(Assumes clock in GHz and length in ns)


    debug("pb_inst_pbonly: inst=%d, inst_data=%d,length=%f\n",inst,inst_data,length);

	if (delay < 2)
	{
        spinerr = "Instruction delay is too small to work with your board";
        debug("pb_inst_pbonly: %s\n", spinerr);
		return -91;
	}

	if(inst == LOOP)
	{
        if(inst_data == 0) {
            spinerr = "Number of loops must be 1 or more";
            debug("pb_inst_pbonly: %s\n", spinerr);        
            return -1;
        }
		inst_data -= 1;
	}
	if(inst == LONG_DELAY)
	{
        if (inst_data == 0 || inst_data == 1) {
            spinerr = "Number of repetitions must be 2 or more";
            debug("pb_inst_pbonly: %s\n", spinerr);        
            return -1;
        }
		inst_data -= 2;
	}

	//-------------------PB CORE counter issue--------------------------------------------------------------------------------------------------------------------------------------------------
	//An extra clock cycle must be subtracted from all instructions that result in a value ending in 0xFF being sent to the core counter (with the exception of 0x0FF).
	//Boards which have been fixed (and which have readable firmware IDs) do not require this and will bypass the following software fix.
	//
	//For boards which have been fixed but do not have firmware registers, there is no way for spinapi to know whether or not the fault has been corrected.
	//Therefore, on all generic PulseBlaster boards and also on the older PBESR boards (boards using AMCC_DEVID and PBESR_PRO_DEVID) the 'FF' fix will
	//be applied by default.  If you wish to bypass this fix, please use:
	//	pb_bypass_FF_fix (1);     //bypass the fix below - no extra clock cycle will be subtracted ( board[cur_board].has_FF_fix will be set to 1 )
	//	pb_bypass_FF_fix(0):	//turn on the software fix - ( board[cur_board].has_FF_fix will be set to 0 )
	//
	if(board[cur_board].has_FF_fix != 1)
	{	
		if( ((delay&0xFF)==0xFF) && (delay > 0xFF) ) 
		{
			delay--;
			debug("pb_inst_pbonly: __ONE CLOCK CYCLE SUBTRACTED__\n", spinerr);  
		}
	}
	//_______________________________________________________________________________________________________________
	
	return pb_inst_direct(flags,inst,inst_data,delay);
}


/**
 * This function allows you to directly specify the fields for an instruction, which will
 * then be passed directly to the board without any modification by software.
 * See your product manual for a discussion of the meaning of each field.
 * <p>
 * This function is provided mainly to help us debug the boards. It is highly 
 * recommended that users make use the higher level instruction functions such 
 * as pb_inst_pbonly(), pb_inst_tworf(), pb_inst_radio(). These allow the
 * creation of instructions in a more user-friendly manner
 * and also perform additional error checking to help you avoid accidentally
 * using paramaters that are out of range for your board.
 *
 * \param flags Flag field (24 bits)
 * \param inst Instruction (4 bits)
 * \param inst_data_direct Instruction data (20 bits) Unlike the other pb_inst*
 * instructions, this field is passed directly to the board and not adjusted based
 * on the instruction used. (eg, using a value of 2 for a loop instruction will cause
 * 3 loops to happen. The other pb_inst* functions would modify this value so the
 * number of loops entered is the number produced)
 * \param length Delay field (32 bits) Note that this is the value is NOT the number
 * of clock cycles an instruction will execute for. There is typically an additional
 * fixed internal delay which is added to produce the actual delay.
 */
int pb_inst_direct(int flags, int inst, int inst_data_direct, int length)
{
    int i;
    int return_value;
    unsigned int temp_byte;
	unsigned int BIT_MASK = 0xFF0000;
	unsigned int OCW = 0;
	unsigned int OPCODE = 0;
	unsigned int DELAY = 0;

    spinerr = noerr;

    // Make sure the different fields of the instruction are within 
    // the proper range
    if(flags != (flags & 0x0FFFFFF)) {
        spinerr = "Flag word is limited to 24 bits";
        debug("pb_inst_direct: %s\n", spinerr);
        return -1;
    }
    if(inst > 8) {
        spinerr = "Invalid opcode";
        debug("pb_inst_direct: %s\n", spinerr);
        return -1;
    }
    if(inst_data_direct != (inst_data_direct & 0x0FFFFF)) {
        spinerr = "Instruction is limited to 20 bits";
        debug("pb_inst_direct: %s\n", spinerr);        
        return -1;
    }

	OPCODE = (inst) | (inst_data_direct << 4);
	OCW = flags;
    DELAY = length;

    debug("pb_inst_direct: OPCODE=0x%x, flags=0x%.6x, delay=%d\n",OPCODE,OCW,DELAY);

	for(i =0; i<3; i++)
	{
		temp_byte = BIT_MASK & OCW;
		temp_byte >>= 16;
		return_value = pb_outp(port_base + 6, temp_byte);
		if(return_value != 0 && (!(ISA_BOARD)))
		{
            spinerr = "Communications error (loop 1)";
            debug("pb_inst_direct: %s\n", spinerr);        
            debug("return value was: %d\n", return_value);
			return return_value;
		}
		OCW <<= 8;
	}
	BIT_MASK = 0xFF0000;
	for(i = 0; i<3; i++)
	{
	
		temp_byte = BIT_MASK & OPCODE;
		temp_byte >>= 16;
		return_value = pb_outp(port_base + 6, temp_byte);
		if(return_value != 0 && (!(ISA_BOARD)))
		{
            spinerr = "Communications error (loop 2)";
            debug("pb_inst_direct: %s\n", spinerr);        
			return return_value;
		}
		OPCODE <<= 8;
	}
	BIT_MASK = 0xFF000000;
	for(i = 0; i<4; i++)
	{
		temp_byte = BIT_MASK & DELAY;
		temp_byte >>= 24;
		return_value = pb_outp(port_base + 6, temp_byte);
		if(return_value != 0 && (!(ISA_BOARD)))
		{
            spinerr = "Communications error (loop 3)";
            debug("pb_inst_direct: %s\n", spinerr);
			return return_value;
		}
		DELAY <<= 8;
	}

	num_instructions +=1;
	return num_instructions - 1;
}

/**
 * Read status from the board. Not all boards support this, see your manual.
 * Each bit of the returned integer indicates whether the board is in that
 * state. Bit 0 is the least significant bit.
 *<ul>
 *<li>Bit 0 - Stopped
 *<li>Bit 1 - Reset
 *<li>Bit 2 - Running
 *<li>Bit 3 - Waiting
 *<li>Bit 4 - Scanning (RadioProcessor boards only)
 *</ul>
 * Bits 5-31 are reserved for future use. It should not be assumed that these
 * will be set to 0.
 * \return Word that indicates the state of the current board as described above.
 */
int pb_read_status()
{
    int status;

    spinerr = noerr;

    if(board[cur_board].status_oldstyle) {
        debug("pb_read_status: using oldstyle\n");
    	/*return pb_inp(0);*/
    	return -1;
    }
    else {
        status = reg_read (REG_EXPERIMENT_RUNNING);
        return status;
    }
}







